import pathlib
import re

import git
import requests
from funcy import collecting, re_find, re_test, silent

from ballet.compat import safepath
from ballet.util import one_or_raise

FILE_CHANGES_COMMIT_RANGE = '{a}...{b}'
REV_REGEX = r'[a-zA-Z0-9_/^@{}-]+'
COMMIT_RANGE_REGEX = re.compile(
    r'(?P<a>{rev})\.\.(?P<thirddot>\.?)(?P<b>{rev})'
    .format(rev=REV_REGEX))
PR_REF_PATH_REGEX = re.compile(r'refs/heads/pull/(\d+)')
GIT_PUSH_FAILURE = (
    git.PushInfo.REJECTED |
    git.PushInfo.REMOTE_REJECTED |
    git.PushInfo.REMOTE_FAILURE |
    git.PushInfo.ERROR
)


class Differ:

    def diff(self):
        a, b = self._get_diff_endpoints()
        return a.diff(b)

    def _get_diff_endpoints(self):
        raise NotImplementedError


class CustomDiffer(Differ):

    def __init__(self, endpoints):
        self.endpoints = endpoints

    def _get_diff_endpoints(self):
        return self.endpoints


class PullRequestBuildDiffer(Differ):
    """Diff files from this pull request against a comparison ref

    Args:
        pr_num (str, int): pull request number
        repo (git.Repo): repo
    """

    def __init__(self, pr_num, repo):
        self.pr_num = int(pr_num)
        self.repo = repo
        self._check_environment()

    def _check_environment(self):
        raise NotImplementedError


class LocalPullRequestBuildDiffer(PullRequestBuildDiffer):

    @property
    def _pr_name(self):
        return self.repo.head.ref.name

    @property
    def _pr_path(self):
        return self.repo.head.ref.path

    def _check_environment(self):
        assert re_test(PR_REF_PATH_REGEX, self._pr_path)

    def _get_diff_endpoints(self):
        a = self.repo.rev_parse('master')
        b = self.repo.rev_parse(self._pr_name)
        return a, b


class LocalMergeBuildDiffer(Differ):
    """Diff files on a merge commit on the
    current active branch. Merge parent order is guaranteed
    such that parent 1 is HEAD and parent 2 is topic[1]

    Attributes:
        repo (git.Repo): The repository to check the merge diff on.
            Must be currently on a branch where the most recent commit
            is a merge.

    References:
        [1] https://git-scm.com/book/en/v2/Git-Tools-Advanced-Merging
    """

    def __init__(self, repo):
        self.repo = repo
        self._check_environment()

    def _check_environment(self):
        assert is_merge_commit(self.repo.head.commit)

    def _get_diff_endpoints(self):
        a = self.repo.head.commit.parents[0]
        b = self.repo.head.commit.parents[1]
        return a, b


def make_commit_range(a, b):
    return FILE_CHANGES_COMMIT_RANGE.format(a=a, b=b)


def get_diff_endpoints_from_commit_range(repo, commit_range):
    """Get endpoints of a diff given a commit range

    The resulting endpoints can be diffed directly::

        a, b = get_diff_endpoints_from_commit_range(repo, commit_range)
        a.diff(b)

    For details on specifying git diffs, see ``git diff --help``.
    For details on specifying revisions, see ``git help revisions``.

    Args:
        repo (git.Repo): Repo object initialized with project root
        commit_range (str): commit range as would be interpreted by ``git
            diff`` command. Unfortunately only patterns of the form ``a..b``
            and ``a...b`` are accepted. Note that the latter pattern finds the
            merge-base of a and b and uses it as the starting point for the
            diff.

    Returns:
        Tuple[git.Commit, git.Commit]: starting commit, ending commit (
            inclusive)

    Raises:
        ValueError: commit_range is empty or ill-formed

    See also:

        <https://stackoverflow.com/q/7251477>
    """
    if not commit_range:
        raise ValueError('commit_range cannot be empty')

    result = re_find(COMMIT_RANGE_REGEX, commit_range)
    if not result:
        raise ValueError(
            'Expected diff str of the form \'a..b\' or \'a...b\' (got {})'
            .format(commit_range))
    a, b = result['a'], result['b']
    a, b = repo.rev_parse(a), repo.rev_parse(b)
    if result['thirddot']:
        a = one_or_raise(repo.merge_base(a, b))
    return a, b


def get_repo(repo=None):
    if repo is None:
        repo = git.Repo(safepath(pathlib.Path.cwd()),
                        search_parent_directories=True)
    return repo


@silent
def get_pr_num(repo=None):
    repo = get_repo(repo)
    pr_num = re_find(PR_REF_PATH_REGEX, repo.head.ref.path)
    return int(pr_num)


@silent
def get_branch(repo=None):
    repo = get_repo(repo)
    branch = repo.head.ref.name
    return branch


def switch_to_new_branch(repo, name):
    new_branch = repo.create_head(name)
    repo.head.ref = new_branch


def is_merge_commit(commit):
    return len(commit.parents) > 1


def set_config_variables(repo, variables):
    """Set config variables

    Args:
        repo (git.Repo): repo
        variables (dict): entries of the form 'user.email': 'you@example.com'
    """
    with repo.config_writer() as writer:
        for k, value in variables.items():
            section, option = k.split('.')
            writer.set_value(section, option, value)
        writer.release()


def get_pull_requests(owner, repo, state='closed'):
    base = 'https://api.github.com'
    q = '/repos/{owner}/{repo}/pulls'.format(owner=owner, repo=repo)
    url = base + q
    headers = {
        'Accept': 'application/vnd.github.v3+json'
    }
    params = {
        'state': state,
        'base': 'master',
        'sort': 'created',
        'direction': 'asc',
    }
    res = requests.get(url, headers=headers, params=params)
    res.raise_for_status()
    return res.json()


@collecting
def get_pull_request_outcomes(owner, repo):
    prs = get_pull_requests(owner, repo, state='closed')
    for pr in prs:
        if pr['merged_at'] is not None:
            yield 'accepted'
        else:
            yield 'rejected'


def did_git_push_succeed(push_info):
    """Check whether a git push succeeded

    A git push succeeded if it was not "rejected" or "remote rejected",
    and if there was not a "remote failure" or an "error".

    Args:
        push_info (git.remote.PushInfo): push info
    """
    return push_info.flags & GIT_PUSH_FAILURE == 0



import os
import pathlib
import pickle

import h5py
import numpy as np
import pandas as pd
from stacklog import stacklog

from ballet.compat import safepath
from ballet.util.fs import splitext2
from ballet.util.log import logger


def _check_ext(ext, expected):
    if ext != expected:
        msg = ('File path has wrong extension: {} (expected {})'
               .format(ext, expected))
        raise ValueError(msg)


def write_tabular(obj, filepath):
    """Write tabular object in HDF5 or pickle format

    Args:
        obj (array or DataFrame): tabular object to write
        filepath (path-like): path to write to; must end in '.h5' or '.pkl'
    """
    _, fn, ext = splitext2(filepath)
    if ext == '.h5':
        _write_tabular_h5(obj, filepath)
    elif ext == '.pkl':
        _write_tabular_pickle(obj, filepath)
    else:
        raise NotImplementedError


def _write_tabular_pickle(obj, filepath):
    _, fn, ext = splitext2(filepath)
    _check_ext(ext, '.pkl')
    if isinstance(obj, np.ndarray):
        with open(safepath(filepath), 'wb') as f:
            pickle.dump(obj, f)
    elif isinstance(obj, pd.core.frame.NDFrame):
        obj.to_pickle(filepath)
    else:
        raise NotImplementedError


def _write_tabular_h5(obj, filepath):
    _, fn, ext = splitext2(filepath)
    _check_ext(ext, '.h5')
    if isinstance(obj, np.ndarray):
        with h5py.File(filepath, 'w') as hf:
            hf.create_dataset(fn, data=obj)
    elif isinstance(obj, pd.core.frame.NDFrame):
        obj.to_hdf(filepath, key=fn)
    else:
        raise NotImplementedError


def read_tabular(filepath):
    """Read tabular object in HDF5 or pickle format

    Args:
        filepath (path-like): path to read to; must end in '.h5' or '.pkl'
    """
    _, fn, ext = splitext2(filepath)
    if ext == '.h5':
        return _read_tabular_h5(filepath)
    elif ext == '.pkl':
        return _read_tabular_pickle(filepath)
    else:
        raise NotImplementedError


def _read_tabular_h5(filepath):
    _, fn, ext = splitext2(filepath)
    _check_ext(ext, '.h5')
    with h5py.File(filepath, 'r') as hf:
        dataset = hf[fn]
        data = dataset[:]
        return data


def _read_tabular_pickle(filepath):
    _, fn, ext = splitext2(filepath)
    _check_ext(ext, '.pkl')
    with open(safepath(filepath), 'rb') as f:
        return pickle.load(f)


def save_model(model, output_dir, name='model'):
    _save_thing(model, output_dir, name,
                savefn=lambda thing, fn: thing.dump(fn))


def save_predictions(ypred, output_dir, name='predictions'):
    """Save predictions to output directory"""
    _save_thing(ypred, output_dir, name)


def save_features(X, output_dir, name='features'):
    """Save built features to output directory"""
    _save_thing(X, output_dir, name)


def save_targets(y, output_dir, name='target'):
    """Save built target to output directory"""
    _save_thing(y, output_dir, name)


def _save_thing(thing, output_dir, name, savefn=write_tabular):
    fn = pathlib.Path(output_dir).joinpath('{}.pkl'.format(name))
    with stacklog(logger.info, 'Saving {} to {}'.format(name, fn)):
        os.makedirs(output_dir, exist_ok=True)
        savefn(thing, fn)


def load_table_from_config(input_dir, config):
    """Load table from table config dict

    Args:
        input_dir (path-like): directory containing input files
        config (dict): mapping with keys 'name', 'path', and 'pd_read_kwargs'.

    Returns:
        pd.DataFrame
    """
    path = pathlib.Path(input_dir).joinpath(config['path'])
    kwargs = config['pd_read_kwargs']
    return pd.read_csv(path, **kwargs)



import importlib
import pathlib
import pkgutil

from ballet.util.log import logger


def import_module_from_modname(modname):
    """Import module"""
    return importlib.import_module(modname)


def import_module_from_relpath(path):
    """Import module at relative path to project root"""
    modname = relpath_to_modname(path)
    return import_module_from_modname(modname)


def import_module_at_path(modname, modpath):
    """Import module from path that may not be on system path

    Args:
        modname (str): module name from package root, e.g. foo.bar
        modpath (str): absolute path to module itself,
            e.g. /home/user/foo/bar.py. In the case of a module that is a
            package, then the path should be specified as '/home/user/foo' and
            a file '/home/user/foo/__init__.py' *must be present* or the import
            will fail.

    Examples:
        >>> modname = 'foo.bar.baz'
        >>> modpath = '/home/user/foo/bar/baz.py'
        >>> import_module_at_path(modname, modpath)
        <module 'foo.bar.baz' from '/home/user/foo/bar/baz.py'>

        >>> modname = 'foo.bar'
        >>> modpath = '/home/user/foo/bar'
        >>> import_module_at_path(modname, modpath)
        <module 'foo.bar' from '/home/user/foo/bar/__init__.py'>
    """
    # TODO just keep navigating up in the source tree until an __init__.py is
    # not found?

    modpath = pathlib.Path(modpath).resolve()

    if modpath.name == '__init__.py':
        # TODO improve debugging output with recommend change
        raise ValueError('Don\'t provide the __init__.py!')

    def is_package(modpath):
        return modpath.suffix != '.py'

    def has_init(dir):
        return dir.joinpath('__init__.py').is_file()

    def has_package_structure(modname, modpath):
        modparts = modname.split('.')
        n = len(modparts)
        dir = modpath
        if not is_package(modpath):
            n = n - 1
            dir = dir.parent
        while n > 0:
            if not has_init(dir):
                return False
            dir = dir.parent
            n = n - 1

        return True

    if not has_package_structure(modname, modpath):
        raise ImportError('Module does not have valid package structure.')

    parentpath = str(pathlib.Path(modpath).parent)
    finder = pkgutil.get_importer(parentpath)
    loader = finder.find_module(modname)
    if loader is None:
        raise ImportError(
            'Failed to find loader for module {} within dir {}'
            .format(modname, parentpath))
    mod = loader.load_module(modname)

    # TODO figure out what to do about this
    assert mod.__name__ == modname

    return mod


def relpath_to_modname(relpath):
    """Convert relative path to module name

    Within a project, a path to the source file is uniquely identified with a
    module name. Relative paths of the form 'foo/bar' are *not* converted to
    module names 'foo.bar', because (1) they identify directories, not regular
    files, and (2) already 'foo/bar/__init__.py' would claim that conversion.

    Args:
        relpath (str): Relative path from some location on sys.path

    Example:
        >>> relpath_to_modname('ballet/util/_util.py')
        'ballet.util._util'
    """
    # don't try to resolve!
    p = pathlib.Path(relpath)

    if p.name == '__init__.py':
        p = p.parent
    elif p.suffix == '.py':
        p = p.with_suffix('')
    else:
        msg = 'Cannot convert a non-python file to a modname'
        msg_detail = 'The relpath given is: {}'.format(relpath)
        logger.error(msg + '\n' + msg_detail)
        raise ValueError(msg)

    return '.'.join(p.parts)


def modname_to_relpath(modname, project_root=None, add_init=True):
    """Convert module name to relative path.

    The project root is usually needed to detect if the module is a package, in
    which case the relevant file is the `__init__.py` within the subdirectory.

    Example:
        >>> modname_to_relpath('foo.features')
        'foo/features.py'
        >>> modname_to_relpath('foo.features',
                               project_root='/path/to/project')
        'foo/features/__init__.py'

    Args:
        modname (str): Module name, e.g. `os.path`
        project_root (str): Path to project root
        add_init (bool): Whether to add `__init__.py` to the path of modules
            that are packages. Defaults to True

    Returns:
        str
    """
    parts = modname.split('.')
    relpath = pathlib.Path(*parts)

    # is the module a package? if so, the relpath identifies a directory
    # it is easier to check for whether a file is a directory than to try to
    # import the module dynamically and see whether it is a package
    if project_root is not None:
        relpath_resolved = pathlib.Path(project_root).joinpath(relpath)
    else:
        relpath_resolved = relpath

    if relpath_resolved.is_dir():
        if add_init:
            relpath = relpath.joinpath('__init__.py')
    else:
        relpath = str(relpath) + '.py'

    return str(relpath)



import unittest
from unittest.mock import patch

from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.validation.feature_acceptance.validator import (
    GFSSFAccepter, NeverAccepter, RandomAccepter)
from tests.util import load_regression_data


class NoOpAccepterTest(unittest.TestCase):

    def test_accepter(self):
        X = None
        y = None
        existing_features = []
        feature = None

        expected = False

        accepter = NeverAccepter(X, y, existing_features, feature)
        actual = accepter.judge()

        self.assertEqual(expected, actual)


class RandomAccepterTest(unittest.TestCase):

    @patch('random.uniform')
    def test_accepter(self, mock_uniform):
        X = None
        y = None
        existing_features = []
        candidate_feature = None

        expected = True

        mock_uniform.return_value = 0.0  # makes sure feature is accepted

        accepter = RandomAccepter(X, y, existing_features, candidate_feature)
        actual = accepter.judge()

        self.assertEqual(expected, actual)


class GFSSFAccepterTest(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_regression_data(n_informative=1,
                                              n_uninformative=14)

    def test_init(self):
        feature_1 = Feature(
            input='A_0',
            transformer=IdentityTransformer(),
            source='1st Feature')
        feature_2 = Feature(
            input='Z_0',
            transformer=IdentityTransformer(),
            source='2nd Feature')

        features = [feature_1]
        candidate_feature = feature_2

        accepter = GFSSFAccepter(self.X, self.y, features, candidate_feature)

        self.assertIsNotNone(accepter)



import traceback
from collections import Counter, namedtuple
from inspect import signature

import numpy as np
import pandas as pd
from funcy import identity, is_seqcont, select_values
from sklearn.base import TransformerMixin
from sklearn_pandas.pipeline import TransformerPipeline

from ballet.eng.base import SimpleFunctionTransformer
from ballet.exc import UnsuccessfulInputConversionError
from ballet.util import DeepcopyMixin, asarray2d, indent, quiet
from ballet.util.log import logger


def make_robust_transformer(transformer):
    if is_seqcont(transformer):
        transformer = map(_replace_callable_with_transformer, transformer)
        map(_validate_transformer_api, transformer)
        return make_robust_transformer_pipeline(*transformer)
    else:
        transformer = _replace_callable_with_transformer(transformer)
        _validate_transformer_api(transformer)
        return DelegatingRobustTransformer(transformer)


def _name_estimators(estimators):
    """Generate names for estimators.

    Adapted from sklearn.pipeline._name_estimators
    """

    def get_name(estimator):
        if isinstance(estimator, DelegatingRobustTransformer):
            return get_name(estimator._transformer)

        return type(estimator).__name__.lower()

    names = list(map(get_name, estimators))
    counter = dict(Counter(names))
    counter = select_values(lambda x: x > 1, counter)

    for i in reversed(range(len(estimators))):
        name = names[i]
        if name in counter:
            names[i] += "-%d" % counter[name]
            counter[name] -= 1

    return list(zip(names, estimators))


def make_transformer_pipeline(*steps):
    """Construct a TransformerPipeline from the given estimators.

    Source: sklearn_pandas.cont_method
    """
    return TransformerPipeline(_name_estimators(steps))


def make_robust_transformer_pipeline(*steps):
    """Construct a transformer pipeline of DelegatingRobustTransformers"""
    steps = list(map(DelegatingRobustTransformer, steps))
    return make_transformer_pipeline(*steps)


ConversionApproach = namedtuple('ConversionApproach', 'name convert caught')


class DelegatingRobustTransformer(DeepcopyMixin, TransformerMixin):
    """Robust transformer that delegates to underlying transformer

    This transformer is robust against different typed and shaped input data.
    It tries a variety of input data conversion approaches and passes the
    result to the underlying transformer, using the first approach that works.

    Args:
        transformer: a transformer object with fit and transform methods

    Raises:
        UnsuccessfulInputConversionError: If none of the conversion approaches
            work.

    """

    DEFAULT_CAUGHT = (ValueError, TypeError)

    CONVERSION_APPROACHES = [
        ConversionApproach('identity', identity, DEFAULT_CAUGHT),
        ConversionApproach('series', pd.Series, DEFAULT_CAUGHT),
        ConversionApproach('dataframe', pd.DataFrame, DEFAULT_CAUGHT),
        ConversionApproach('array', np.asarray, DEFAULT_CAUGHT),
        ConversionApproach('asarray2d', asarray2d, ()),
    ]

    def __init__(self, transformer):
        self._transformer = transformer
        self._stored_conversion_approach = None

    def __getattr__(self, attr):
        if '_transformer' in self.__dict__:
            return getattr(self._transformer, attr)
        else:
            raise AttributeError

    def __getstate__(self):
        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __repr__(self):
        name = type(self).__name__
        return '{name}({transformer!r})'.format(
            name=name, transformer=self._transformer)

    @property
    def _tname(self):
        return type(self._transformer).__name__

    def fit(self, X, y=None, **kwargs):
        # don't return the result of transformer.fit because it is the
        # underlying transformer, not this robust transformer
        self._call_robust(self._transformer.fit, X, y, kwargs)

        # instead, return this robust transformer
        return self

    def transform(self, X, y=None, **kwargs):
        return self._call_robust(self._transformer.transform, X, y, kwargs)

    @staticmethod
    def _call_with_convert(method, convert, X, y, kwargs):
        if y is not None:
            return method(convert(X), y=convert(y), **kwargs)
        else:
            return method(convert(X), **kwargs)

    @quiet
    def _call_robust(self, method, X, y, kwargs):
        if self._stored_conversion_approach is not None:
            approach = self._stored_conversion_approach
            self._log_attempt_using_stored_approach(approach)
            convert = approach.convert
            try:
                result = self._call_with_convert(method, convert, X, y, kwargs)
                self._log_success_using_stored_approach(approach)
                return result
            except Exception as e:
                self._log_failure_using_stored_approach(approach, e)
                raise
        else:
            for approach in DelegatingRobustTransformer.CONVERSION_APPROACHES:
                try:
                    self._log_attempt(approach)
                    result = self._call_with_convert(
                        method, approach.convert, X, y, kwargs)
                    self._log_success(approach)
                    self._stored_conversion_approach = approach
                    return result
                except approach.caught as e:
                    self._log_catch(approach, e)
                    continue
                except Exception as e:
                    self._log_error(approach, e)
                    raise

            self._log_failure_no_more_approaches()
            raise UnsuccessfulInputConversionError

    def _log_attempt_using_stored_approach(self, approach):
        logger.debug(
            '{tname}: '
            'Attempting to convert using stored, '
            'previously-successful approach {approach.name!r}'
            .format(tname=self._tname, approach=approach))

    def _log_failure_using_stored_approach(self, approach, e):
        pretty_tb = self._get_pretty_tb()
        exc_name = type(e).__name__
        logger.debug(
            '{tname}: '
            'Conversion unexpectedly failed using stored, '
            'previously-successful approach {approach.name!r} '
            'because of error {exc_name!r}\n\n{tb}'
            .format(tname=self._tname, approach=approach, exc_name=exc_name,
                    tb=pretty_tb))

    def _log_success_using_stored_approach(self, approach):
        logger.debug(
            '{tname}: '
            'Conversion with stored, previously-successful approach '
            '{approach.name!r} succeeded!'
            .format(tname=self._tname, approach=approach))

    def _log_attempt(self, approach):
        logger.debug(
            '{tname}: '
            'Attempting to convert using approach {approach.name!r}...'
            .format(tname=self._tname, approach=approach))

    def _get_pretty_tb(self):
        tb = traceback.format_exc()
        pretty_tb = indent(tb, n=8)
        return pretty_tb

    def _log_catch(self, approach, e):
        pretty_tb = self._get_pretty_tb()
        exc_name = type(e).__name__
        logger.debug(
            '{tname}: '
            'Conversion approach {approach.name!r} didn\'t work, '
            'caught exception {exc_name!r}\n\n{tb}'
            .format(tname=self._tname, approach=approach, exc_name=exc_name,
                    tb=pretty_tb))

    def _log_error(self, approach, e):
        pretty_tb = self._get_pretty_tb()
        exc_name = type(e).__name__
        logger.debug(
            '{tname}: '
            'Conversion failed during {approach.name!r} because of '
            'an unrecoverable error {exc_name!r}\n\n{tb}'
            .format(tname=self._tname, approach=approach, exc_name=exc_name,
                    tb=pretty_tb))

    def _log_success(self, approach):
        logger.debug(
            '{tname}: '
            'Conversion approach {approach.name!r} succeeded!'
            .format(tname=self._tname, approach=approach))

    def _log_failure_no_more_approaches(self):
        logger.info('Conversion failed, and we\'re not sure why...')


def _validate_transformer_api(transformer):
    if not all(
        hasattr(transformer, attr)
        for attr in ['fit', 'transform', 'fit_transform']
    ):
        raise ValueError('Transformer object missing required attribute')

    sig_fit = signature(transformer.fit)
    if '(X, y=None' not in str(sig_fit):
        raise ValueError(
            'Invalid signature for transformer.fit: {sig_fit}'
            .format(sig_fit=sig_fit))

    sig_transform = signature(transformer.transform)
    if '(X' not in str(sig_transform):
        raise ValueError(
            'Invalid signature for transformer.transform: {sig_transform}'
            .format(sig_fit=sig_fit))


def _replace_callable_with_transformer(transformer):
    if callable(transformer) and not isinstance(transformer, type):
        return SimpleFunctionTransformer(transformer)
    else:
        return transformer



import os
import os.path
import pathlib
from shutil import copyfile

from funcy import partial, suppress

from ballet.compat import safepath
from ballet.exc import BalletError
from ballet.util.log import logger


def spliceext(filepath, s):
    """Add s into filepath before the extension

    Args:
        filepath (PathLike): file path
        s (str): string to splice

    Returns:
        str
    """
    root, ext = os.path.splitext(safepath(filepath))
    return root + s + ext


def replaceext(filepath, new_ext):
    """Replace any existing file extension with a new one

    If the new extension is the empty string, all existing extensions will
    be removed.

    Example::

        >>> replaceext('/foo/bar.txt', 'py')
        '/foo/bar.py'
        >>> replaceext('/foo/bar.txt', '.doc')
        '/foo/bar.doc'

    Args:
        filepath (PathLike): file path
        new_ext (str): new file extension; if a leading dot is not included,
            it will be added.

    Returns:
        str
    """
    if new_ext and not new_ext.startswith('.'):
        new_ext = '.' + new_ext
    return str(pathlib.Path(filepath).with_suffix(new_ext))


def splitext2(filepath):
    """Split filepath into root, filename, ext

    Args:
        filepath (PathLike): file path

    Returns:
        str
    """
    root, filename = os.path.split(safepath(filepath))
    filename, ext = os.path.splitext(safepath(filename))
    return root, filename, ext


def isemptyfile(filepath):
    """Determine if the file both exists and isempty

    Args:
        filepath (PathLike): file path

    Returns:
        bool
    """
    exists = os.path.exists(safepath(filepath))
    if exists:
        filesize = os.path.getsize(safepath(filepath))
        return filesize == 0
    else:
        return False


def synctree(src, dst, onexist=None):
    """Recursively sync files at directory src to dst

    This is more or less equivalent to::

       cp -n -R ${src}/ ${dst}/

    If a file at the same path exists in src and dst, it is NOT overwritten
    in dst. Pass ``onexist`` in order to raise an error on such conditions.

    Args:
        src (path-like): source directory
        dst (path-like): destination directory, does not need to exist
        onexist (callable): function to call if file exists at destination,
            takes the full path to destination file as only argument

    Returns:
        List[Tuple[PathLike,str]]: changes made by synctree, list of tuples of
        the form ("/absolute/path/to/file", "<kind>") where the change kind is
        one of "dir" (new directory was created) or "file" (new file was
        created).
    """
    src = pathlib.Path(src).resolve()
    dst = pathlib.Path(dst).resolve()

    if not src.is_dir():
        raise ValueError

    if dst.exists() and not dst.is_dir():
        raise ValueError

    if onexist is None:
        def onexist(): pass

    return _synctree(src, dst, onexist)


def _synctree(src, dst, onexist):
    result = []
    cleanup = []
    try:
        for root, dirnames, filenames in os.walk(safepath(src)):
            root = pathlib.Path(root)
            relative_dir = root.relative_to(src)

            for dirname in dirnames:
                dstdir = dst.joinpath(relative_dir, dirname)
                if dstdir.exists():
                    if not dstdir.is_dir():
                        raise BalletError
                else:
                    logger.debug(
                        'Making directory: {dstdir!s}'.format(dstdir=dstdir))
                    dstdir.mkdir()
                    result.append((dstdir, 'dir'))
                    cleanup.append(partial(os.rmdir, safepath(dstdir)))

            for filename in filenames:
                srcfile = root.joinpath(filename)
                dstfile = dst.joinpath(relative_dir, filename)
                if dstfile.exists():
                    onexist(dstfile)
                else:
                    logger.debug(
                        'Copying file to destination: {dstfile!s}'
                        .format(dstfile=dstfile))
                    copyfile(srcfile, dstfile)
                    result.append((dstfile, 'file'))
                    cleanup.append(partial(os.unlink, safepath(dstfile)))

    except Exception:
        with suppress(Exception):
            for f in reversed(cleanup):
                f()
        raise

    return result


def pwalk(d, **kwargs):
    """Similar to os.walk but with pathlib.Path objects

    Returns:
        Iterable[Path]
    """
    for dirpath, dirnames, filenames in os.walk(safepath(d), **kwargs):
        dirpath = pathlib.Path(dirpath)
        for p in dirnames + filenames:
            yield dirpath.joinpath(p)



import unittest
from unittest.mock import patch

import numpy as np

from ballet.util import asarray2d
from ballet.util.log import logger
from ballet.util.testing import ArrayLikeEqualityTestingMixin
from ballet.validation.entropy import (
    NEIGHBORS_ALGORITHM, NEIGHBORS_METRIC, _compute_empirical_probability,
    _compute_epsilon, _compute_n_points_within_radius_i,
    _compute_volume_unit_ball, _estimate_cont_entropy, _estimate_disc_entropy,
    _is_column_cont, _is_column_disc, _make_neighbors,
    estimate_conditional_information, estimate_entropy,
    estimate_mutual_information, nonnegative)


class EntropyTest(ArrayLikeEqualityTestingMixin, unittest.TestCase):

    def test_nonnegative_positive_output(self):
        @nonnegative()
        def func():
            return 1

        self.assertEqual(1, func())

    def test_nonnegative_negative_output(self):
        @nonnegative(name="Result")
        def func():
            return -1

        self.assertEqual(0, func())

    def test_nonnegative_negative_introspection(self):
        @nonnegative()
        def estimate_something():
            return -1

        with self.assertLogs(logger.name, 'WARNING') as cm:
            estimate_something()
        self.assertEqual(1, len(cm.output))
        self.assertIn("Something", cm.output[0])

    def test_make_neighbors(self):
        nn = _make_neighbors()
        self.assertEqual(NEIGHBORS_ALGORITHM, nn.algorithm)
        self.assertEqual(NEIGHBORS_METRIC, nn.metric)

    def test_compute_nx_i(self):
        # Note the chebyshev distance is used
        n = 5
        x = np.array([
            [0, 0],
            [1, 0],
            [0, 1],
            [1, 1],
            [0.5, 0.5],
        ])
        radius = np.array([
            [0.7],
            [10],
            [1],
            [0.7],
            [0.1],
        ])
        expected_nx = np.array([
            2,  # 0,0 and 0.5, 0.5
            5,  # all points
            2,  # would be all points, but should specifically exclude
                # points on the margin
            2,  # 1, 1 and 0.5, 0.5
            1   # just 0.5, 0.5
        ])

        nn = _make_neighbors().fit(x)
        for i in range(n):
            x_i = x[i:i + 1, :]  # require a (1,m) row array
            radius_i = radius[i]
            nx_i = _compute_n_points_within_radius_i(nn, x_i, radius_i)
            expected_nx_i = expected_nx[i]
            self.assertEqual(expected_nx_i, nx_i)

    def test_compute_empirical_probability(self):
        x = [1, 1, 2, 3, 2, 1, 1, 2]
        expected_pk = np.array([4 / 8, 3 / 8, 1 / 8])
        expected_events = np.array([[1], [2], [3]])
        pk, events = _compute_empirical_probability(x)

        self.assertArrayEqual(expected_pk, pk)
        self.assertArrayEqual(expected_events, events)

    def test_compute_volume_unit_ball_chebyshev(self):
        metric = 'chebyshev'
        expected_volume = 1
        for d in [1, 2, 5, 11]:
            volume = _compute_volume_unit_ball(d, metric=metric)
            self.assertEqual(expected_volume, volume)

    def test_compute_volume_unit_ball_euclidean(self):
        metric = 'euclidean'
        volume_upper_bound = 1
        for d in [1, 2, 5, 11]:
            volume = _compute_volume_unit_ball(d, metric=metric)
            self.assertLessEqual(volume, volume_upper_bound)

    def test_compute_epsilon(self):
        # data looks like this:
        # |         x
        # |       x
        # |     x
        # |   x
        # | x
        # |---------|
        #
        x = np.array([
            [0.5, 0.5],
            [1.5, 1.5],
            [2.5, 2.5],
            [3.5, 3.5],
            [4.5, 4.5],
        ])

        # note k is 3 and distance is chebyshev
        expected_epsilon = np.array([
            [2 * 3.0],
            [2 * 2.0],
            [2 * 2.0],  # has two neighbors at distance 2
            [2 * 2.0],
            [2 * 3.0],
        ])

        epsilon = _compute_epsilon(x)

        self.assertArrayEqual(expected_epsilon, epsilon)

    def test_disc_entropy_constant_vals_1d(self):
        """If x (column vector) is constant, then H(x) = 0"""
        same_val_arr_ones = np.ones((50, 1))
        H_hat = _estimate_disc_entropy(same_val_arr_ones)
        self.assertEqual(0, H_hat)

    def test_disc_entropy_constant_vals_2d(self):
        """If each column in x (matrix), then H(x) = 0"""
        same_val_arr_zero = np.zeros((50, 1))
        same_val_arr_ones = np.ones((50, 1))
        same_multi_val_arr = np.concatenate(
            (same_val_arr_ones, same_val_arr_zero), axis=1)
        H_hat = _estimate_disc_entropy(same_multi_val_arr)
        self.assertEqual(0, H_hat)

    def test_disc_entropy_two_values(self):
        """Entropy of fair coin ~= log(2)"""
        same_val_arr_zero = np.zeros((50, 1))
        same_val_arr_ones = np.ones((50, 1))
        diff_val_arr = np.concatenate(
            (same_val_arr_ones, same_val_arr_zero), axis=0)

        expected_h = np.log(2)
        H_hat = _estimate_disc_entropy(diff_val_arr)
        self.assertAlmostEqual(expected_h, H_hat)

    def test_is_column_disc(self):
        x = asarray2d(np.arange(50))
        result = _is_column_disc(x)
        self.assertTrue(result)

    def test_is_column_cont(self):
        x = asarray2d(np.random.rand(50))
        result = _is_column_cont(x)
        self.assertTrue(result)

    @unittest.skip
    @patch('ballet.validation.entropy._get_disc_columns')
    def test_cont_disc_entropy_differs_disc(self, get_disc_columns):
        """Expect cont, disc columns to have different entropy"""
        disc = asarray2d(np.arange(50))

        # we run into trouble here because as disc as *actually* discrete,
        # epsilon would not be calculated (it is set to some dummy value of
        # -inf). instead, we patch get_disc_columns and "force" epsilon to be
        # calculated
        epsilon = _compute_epsilon(disc)
        H_cont = _estimate_cont_entropy(disc, epsilon)

        H_disc = _estimate_disc_entropy(disc)

        self.assertNotEqual(H_cont, H_disc)

    def test_cont_disc_entropy_differs_cont(self):
        """Expect cont, disc columns to have different entropy"""
        cont = asarray2d(np.arange(50)) + 0.5
        epsilon = _compute_epsilon(cont)

        H_cont = _estimate_cont_entropy(cont, epsilon)
        H_disc = _estimate_disc_entropy(cont)

        self.assertNotEqual(H_cont, H_disc)

    def test_entropy_multiple_disc(self):
        same_val_arr_zero = np.zeros((50, 1))
        same_val_arr_ones = np.ones((50, 1))
        # The 0.5 forces float => classified as continuous
        cont_val_arange = asarray2d(np.arange(50) + 0.5)
        all_disc_arr = np.concatenate(
            (same_val_arr_ones, same_val_arr_zero), axis=1)
        mixed_val_arr = np.concatenate(
            (all_disc_arr, cont_val_arange), axis=1)

        all_disc_h = estimate_entropy(all_disc_arr)
        mixed_h = estimate_entropy(mixed_val_arr)
        self.assertGreater(
            mixed_h,
            all_disc_h,
            msg='Expected adding continuous column increases entropy')

    def test_mi_uninformative(self):
        x = np.reshape(np.arange(1, 101), (-1, 1))
        y = np.ones((100, 1))
        mi = estimate_mutual_information(x, y)
        h_z = estimate_entropy(x)
        self.assertGreater(
            h_z / 4,
            mi,
            'uninformative column should have no information')

    def test_mi_informative(self):
        x = np.reshape(np.arange(1, 101), (-1, 1))
        y = np.reshape(np.arange(1, 101), (-1, 1))
        mi = estimate_mutual_information(x, y)
        h_y = estimate_entropy(y)
        self.assertGreater(
            mi,
            h_y / 4,
            'exact copy columns should have high information')

    def test_cmi_high_info_uninformative_z(self):
        # redundant copies have little information
        x = np.reshape(np.arange(1, 101), (-1, 1))
        y = np.reshape(np.arange(1, 101), (-1, 1))

        # exact copies of y should have lots of information
        useless_z = np.ones((100, 1))
        cmi = estimate_conditional_information(x, y, useless_z)
        mi = estimate_mutual_information(x, y)
        self.assertAlmostEqual(
            cmi,
            mi,
            'uninformative z should not affect mutual information score')

    def test_cmi_redundant_info(self):
        x = np.reshape(np.arange(1, 101), (-1, 1))
        y = np.reshape(np.arange(1, 101), (-1, 1))
        exact_z = np.reshape(np.arange(1, 101), (-1, 1))

        h_y = estimate_entropy(y)
        cmi = estimate_conditional_information(x, y, exact_z)
        self.assertGreater(
            h_y / 4,
            cmi,
            'redundant copies should have little information')



import pathlib
import shutil
import tempfile
from collections import namedtuple
from unittest.mock import patch

import pytest
from cookiecutter.utils import work_in

import ballet
from ballet.compat import safepath
from ballet.project import Project
from ballet.templating import render_project_template
from tests.util import tree


@pytest.fixture
def tempdir():
    """Tempdir fixture using tempfile.TemporaryDirectory"""
    with tempfile.TemporaryDirectory() as d:
        yield pathlib.Path(d)


@pytest.fixture
def quickstart(tempdir):
    """
    $ cd tempdir
    $ ballet quickstart
    $ tree .
    """
    # cd tempdir
    with work_in(safepath(tempdir)):

        project_slug = 'foo-bar'
        package_slug = 'foo_bar'
        extra_context = {
            'project_name': project_slug.capitalize(),
            'project_slug': project_slug,
            'package_slug': package_slug,
        }

        # ballet quickstart
        render_project_template(no_input=True,
                                extra_context=extra_context,
                                output_dir=safepath(tempdir))

        # tree .
        tree(tempdir)

        project = Project.from_path(tempdir.joinpath(project_slug))
        repo = project.repo

        yield (
            namedtuple('Quickstart',
                       'project tempdir project_slug package_slug repo')
            ._make((project, tempdir, project_slug, package_slug, repo))
        )


@pytest.fixture
def project_template_copy(tempdir):
    old_path = ballet.templating.PROJECT_TEMPLATE_PATH
    new_path = tempdir.joinpath('templates', 'project_template')
    shutil.copytree(safepath(old_path), safepath(new_path))

    with patch('ballet.templating.PROJECT_TEMPLATE_PATH', new_path):
        tree(new_path)
        yield new_path



# -*- coding: utf-8 -*-

"""Top-level package for {{ cookiecutter.project_slug }}."""

__author__ = '{{ cookiecutter.full_name.replace("\'", "\\\'") }}'
__email__ = '{{ cookiecutter.email }}'

import {{ cookiecutter.package_slug }}.load_data
import {{ cookiecutter.package_slug }}.features



import unittest
from unittest.mock import MagicMock, patch

import ballet.util.log
from ballet.exc import SkippedValidationTest
from ballet.validation.main import (  # noqa F401
    _check_project_structure, _evaluate_feature_performance, _load_class,
    _prune_existing_features, _validate_feature_api, validate,
    validation_stage)


class MainTest(unittest.TestCase):

    def test_validation_stage_success(self):
        @validation_stage('do something')
        def call():
            return True

        call()

    def test_validation_stage_skipped(self):
        @validation_stage('do something')
        def call():
            raise SkippedValidationTest

        logger = ballet.util.log.logger
        with self.assertLogs(logger=logger) as cm:
            call()

        output = '\n'.join(cm.output)
        self.assertIn('SKIPPED', output)

    def test_validation_stage_failure(self):
        exc = ValueError
        @validation_stage('do something')
        def call():
            raise exc

        with self.assertRaises(exc):
            call()

    @patch('ballet.validation.main.import_module_from_modname')
    def test_load_class(self, mock_import):

        import ballet.validation.project_structure.validator
        from ballet.validation.project_structure.validator import \
            ProjectStructureValidator
        mock_import.return_value = \
            ballet.validation.project_structure.validator

        mock_project = MagicMock()
        modname = 'ballet.validation.project_structure.validator'
        clsname = 'ProjectStructureValidator'
        path = modname + '.' + clsname
        mock_project.config.get.return_value = path

        config_key = None

        logger = ballet.util.log.logger
        with self.assertLogs(logger=logger, level='DEBUG') as cm:
            cls = _load_class(mock_project, config_key)

        output = '\n'.join(cm.output)
        self.assertIn(clsname, output)
        self.assertIs(cls, ProjectStructureValidator)

    @unittest.expectedFailure
    def test_check_project_structure(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_validate_feature_api(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_evaluate_feature_performance(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_prune_existing_features(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_validate(self):
        raise NotImplementedError



import pathlib
import unittest
from unittest.mock import patch

from ballet.templating import (
    _stringify_path, render_feature_template, render_project_template)


class TemplatingTest(unittest.TestCase):

    def test__stringify_path_pathlike(self):
        obj = pathlib.Path()
        output = _stringify_path(obj)
        self.assertIsInstance(output, str)

    def test__stringify_path_not_pathlike(self):
        obj = object()
        output = _stringify_path(obj)
        self.assertNotIsInstance(output, str)

    @patch('ballet.templating.cookiecutter')
    def test_render_project_template(self, mock_cookiecutter):
        render_project_template()

        args, _ = mock_cookiecutter.call_args

        self.assertEqual(len(args), 1)
        path = args[0]
        self.assertIn('project_template', str(path))

    @patch('ballet.templating.cookiecutter')
    def test_render_feature_template(self, mock_cookiecutter):
        render_feature_template()

        args, _ = mock_cookiecutter.call_args

        self.assertEqual(len(args), 1)
        path = args[0]
        self.assertIn('feature_template', str(path))



from ballet.project import config
from ballet.util.io import load_table_from_config
from funcy import some, where

def load_data(input_dir=None):
    """Load data"""
    if input_dir is not None:
        tables = config.get('data.tables')

        entities_table_name = config.get('data.entities_table_name')
        entities_config = some(where(tables, name=entities_table_name))
        X = load_table_from_config(input_dir, entities_config)

        targets_table_name = config.get('data.targets_table_name')
        targets_config = some(where(tables, name=targets_table_name))
        y = load_table_from_config(input_dir, targets_config)
    else:
        raise NotImplementedError

    return X, y



from ballet.project import Project
from ballet.validation.common import subsample_data_for_validation
from ballet.validation.main import _load_class


def validate_feature_acceptance(feature, X, y, subsample=False, path=None,
                                package=None):
    if package is not None:
        project = Project(package)
    elif path is not None:
        project = Project.from_path(path)
    else:
        project = Project.from_cwd()

    if subsample:
        X, y = subsample_data_for_validation(X, y)

    # build project
    result = project.build(X, y)

    # load accepter for this project
    Accepter = _load_class(project, 'validation.feature_accepter')
    accepter = Accepter(result.X_df, result.y, result.features, feature)
    return accepter.judge()



import unittest

import numpy as np
import pandas as pd

import ballet.eng.ts
from ballet.util.testing import ArrayLikeEqualityTestingMixin


class TestTimeSeriesTransformers(
        unittest.TestCase, ArrayLikeEqualityTestingMixin):

    def test_single_lagger(self):
        # simple test
        data = pd.util.testing.makeTimeSeries()

        trans = ballet.eng.ts.SingleLagger(1)
        result = trans.fit_transform(data)
        expected_result = data.shift(1)

        self.assertSeriesEqual(result, expected_result)

        data = pd.DataFrame(
            data={
                'city': ['LA', 'LA', 'LA', 'NYC', 'BOS', 'BOS', 'BOS'],
                'year': [2001, 2002, 2003, 2002, 2003, 2004, 2005],
                'width': [1, 2, 3, 4, 5, 6, 7],
            }
        ).set_index(['city', 'year']).sort_index()
        trans = ballet.eng.ts.SingleLagger(
            1, groupby_kwargs={'level': 'city'})
        result = trans.fit_transform(data)
        expected_result = pd.DataFrame(
            data={
                'city': ['LA', 'LA', 'LA', 'NYC', 'BOS', 'BOS', 'BOS'],
                'year': [2001, 2002, 2003, 2002, 2003, 2004, 2005],
                'width': [np.nan, 1, 2, np.nan, np.nan, 5, 6],
            }
        ).set_index(['city', 'year']).sort_index()

        self.assertFrameEqual(result, expected_result)

    @unittest.expectedFailure
    def test_multi_lagger(self):
        # TODO
        raise AssertionError



import random
from unittest.util import _common_shorten_repr

import funcy
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.util.testing as pdt

EPSILON = 1e-4


@funcy.decorator
def withfailure(call, format):
    try:
        call()
    except AssertionError:
        standardMsg = format.format(
            *_common_shorten_repr(call.first, call.second))
        msg = call.self._formatMessage(call.msg, standardMsg)
        raise call.self.fail(msg)


class ArrayLikeEqualityTestingMixin:
    """Mix-in to enable comparisons of array-like objects in unit tests"""

    @withfailure('{} != {}')
    def assertArrayEqual(self, first, second, msg=None):
        """Test that arrays first and second are equal"""
        npt.assert_array_equal(first, second, verbose=False)

    @withfailure('{} == {}')
    def assertArrayNotEqual(self, first, second, msg=None):
        """Test that arrays first and second are not equal"""
        try:
            npt.assert_array_equal(first, second, verbose=False)
        except AssertionError:
            pass
        else:
            raise AssertionError

    @withfailure('{} !≈ {}')
    def assertArrayAlmostEqual(
            self, first, second, places=6, msg=None, delta=None):
        """Test that arrays first and second are almost equal

        There are two ways to control the tolerance of the approximate
        equality test, using ``places`` or ``delta``. All values are compared
        element-wise.
        ``delta``
        is provided, the allowed tolerance between the values is set to
        ``delta. Else if ``places`` is provided, values are compared up to
        ``places`` places. Else, ``places`` defaults to 6.

        Args:
            first: first object
            second: second object
            places (int, default=6): elements are compared up to this many
                places
            delta (float): the allowed tolerance between elements
        """

        # convert between both "places" and "delta" allowed to just "decimal"
        # numpy says abs(desired-actual) < 1.5 * 10**(-decimal)
        # => delta = 1.5 * 10**(-decimal)
        # => delta / 1.5 = 10**(-decimal)
        # => log10(delta/1.5) = -decimal
        # => decimal = -log10(delta) + log10(1.5)
        if delta is not None:
            places = int(-np.log10(delta) + np.log10(1.5))

        npt.assert_array_almost_equal(
            first, second, decimal=places, verbose=False)

    @withfailure('{} ≈ {}')
    def assertArrayNotAlmostEqual(
            self, first, second, places=6, msg=None, delta=None):
        """Test that arrays first and second are not almost equal"""

        if delta is not None:
            places = int(-np.log10(delta) + np.log10(1.5))

        try:
            npt.assert_array_almost_equal(
                first, second, decimal=places, verbose=False)
        except AssertionError:
            pass
        else:
            raise AssertionError

    def assertFrameEqual(self, first, second, msg=None, **kwargs):
        """Test that DataFrames first and second are equal"""
        self._assertPandasEqual(
            pdt.assert_frame_equal, first, second, msg=msg, **kwargs)

    def assertFrameNotEqual(self, first, second, msg=None, **kwargs):
        """Test that DataFrames first and second are not equal"""
        self._assertPandasNotEqual(
            pdt.assert_frame_equal, first, second, msg=msg, **kwargs)

    def assertSeriesEqual(self, first, second, msg=None, **kwargs):
        """Test that Series first and second are equal"""
        self._assertPandasEqual(
            pdt.assert_series_equal, first, second, msg=msg, **kwargs)

    def assertSeriesNotEqual(self, first, second, msg=None, **kwargs):
        """Test that Series first and second are not equal"""
        self._assertPandasNotEqual(
            pdt.assert_series_equal, first, second, msg=msg, **kwargs)

    def assertIndexEqual(self, first, second, msg=None, **kwargs):
        """Test that Index first and second are equal"""
        self._assertPandasEqual(
            pdt.assert_index_equal, first, second, msg=msg, **kwargs)

    def assertIndexNotEqual(self, first, second, msg=None, **kwargs):
        """Test that Index first and second are not equal"""
        self._assertPandasNotEqual(
            pdt.assert_index_equal, first, second, msg=msg, **kwargs)

    def assertPandasObjectEqual(self, first, second, msg=None, **kwargs):
        """Test that arbitrary Pandas objects first and second are equal"""
        is_pdobj = funcy.isa(pd.core.base.PandasObject)
        if is_pdobj(first) and is_pdobj(second):
            if isinstance(first, type(second)):
                if isinstance(first, pd.DataFrame):
                    self.assertFrameEqual(first, second, msg=msg, **kwargs)
                elif isinstance(first, pd.Series):
                    self.assertSeriesEqual(first, second, msg=msg, **kwargs)
                elif isinstance(first, pd.Index):
                    self.assertIndexEqual(first, second, msg=msg, **kwargs)
                else:
                    # unreachable?
                    raise AssertionError('you found a bug: unreachable code')
        else:
            standardMsg = '{} and {} are uncomparable types'.format(
                *_common_shorten_repr(first, second))
            msg = self._formatMessage(msg, standardMsg)
            raise self.failureException(msg)

    def assertPandasObjectNotEqual(self, first, second, msg=None, **kwargs):
        """Test that arbitrary Pandas objects first and second are not equal"""
        is_pdobj = funcy.isa(pd.core.base.PandasObject)
        if is_pdobj(first) and is_pdobj(second):
            if isinstance(first, type(second)):
                if isinstance(first, pd.DataFrame):
                    self.assertFrameNotEqual(first, second, msg=msg, **kwargs)
                elif isinstance(first, pd.Series):
                    self.assertSeriesNotEqual(first, second, msg=msg, **kwargs)
                elif isinstance(first, pd.Index):
                    self.assertIndexNotEqual(first, second, msg=msg, **kwargs)
                else:
                    # unreachable?
                    raise AssertionError('you found a bug: unreachable code')
        else:
            # it's great that they are uncomparable types :)
            pass

    @withfailure('{} != {}')
    def _assertPandasEqual(self, func, first, second, msg=None, **kwargs):
        func(first, second, **kwargs)

    @withfailure('{} == {}')
    def _assertPandasNotEqual(self, func, first, second, msg=None, **kwargs):
        try:
            func(first, second, **kwargs)
        except AssertionError:
            pass
        else:
            raise AssertionError


@funcy.contextmanager
def seeded(seed):
    """Set seed, run code, then restore rng state"""
    if seed is not None:
        np_random_state = np.random.get_state()
        random_state = random.getstate()
        np.random.seed(seed)
        random.seed(seed)

    yield

    if seed is not None:
        np.random.set_state(np_random_state)
        random.setstate(random_state)


@funcy.contextmanager
def log_seed_on_error(logger, seed=None):
    """Store seed, run code, and report seed if error"""
    if seed is None:
        seed = random.randint(0, 2**32 - 1)
    try:
        with seeded(seed):
            yield
    except Exception:
        logger.exception('Error was thrown using seed {}'.format(seed))



from abc import ABCMeta, abstractmethod

from funcy import constantly, ignore, post_processing


class BaseValidator(metaclass=ABCMeta):
    """Base class for a generic validator"""

    @abstractmethod
    def validate(self):
        """Validate something

        Returns:
            bool: validation succeeded
        """
        pass


class FeaturePerformanceEvaluator(metaclass=ABCMeta):
    """Evaluate the performance of features from an ML point-of-view"""

    def __init__(self, X_df, y, features, candidate_feature):
        self.X_df = X_df
        self.y = y
        self.features = features
        self.candidate_feature = candidate_feature

    def __str__(self):
        return self.__class__.__name__


class FeatureAcceptanceMixin(metaclass=ABCMeta):

    @abstractmethod
    def judge(self):
        """Judge whether feature should be accepted

        Returns:
            bool: feature should be accepted
        """
        pass


class FeaturePruningMixin(metaclass=ABCMeta):

    @abstractmethod
    def prune(self):
        """Prune existing features

        Returns:
            list: list of features to remove
        """
        pass


class FeatureAccepter(FeatureAcceptanceMixin, FeaturePerformanceEvaluator):
    """Accept/reject a feature to the project based on its performance"""
    pass


class FeaturePruner(FeaturePruningMixin, FeaturePerformanceEvaluator):
    """Prune features after acceptance based on their performance"""
    pass


class BaseCheck(metaclass=ABCMeta):

    @ignore(Exception, default=False)
    @post_processing(constantly(True))
    def do_check(self, obj):
        return self.check(obj)

    @abstractmethod
    def check(self, obj):
        pass



import unittest
from textwrap import dedent
from unittest.mock import patch

from ballet.util.ci import TravisPullRequestBuildDiffer
from ballet.util.git import make_commit_range
from ballet.validation.common import ChangeCollector

from .util import (
    SampleDataMixin, make_mock_project, mock_feature_api_validator,
    mock_file_change_validator, null_change_collector)
from ..util import make_mock_commits, mock_repo


class _CommonSetup(SampleDataMixin):

    def setUp(self):
        super().setUp()
        self.pr_num = 73

        self.valid_feature_str = dedent(
            '''
            from sklearn.base import BaseEstimator, TransformerMixin
            class IdentityTransformer(BaseEstimator, TransformerMixin):
                def fit(self, X, y=None, **fit_kwargs):
                    return self
                def transform(self, X, **transform_kwargs):
                    return X
            input = 'size'
            transformer = IdentityTransformer()
            '''
        ).strip()
        self.invalid_feature_str = dedent(
            '''
            from sklearn.base import BaseEstimator, TransformerMixin
            class RaisingTransformer(BaseEstimator, TransformerMixin):
                def fit(self, X, y=None, **fit_kwargs):
                    raise RuntimeError
                def transform(self, X, **transform_kwargs):
                    raise RuntimeError
            input = 'size'
            transformer = RaisingTransformer()
            '''
        ).strip()


class ChangeCollectorTest(_CommonSetup, unittest.TestCase):

    def test_init(self):
        with null_change_collector(self.pr_num) as change_collector:
            self.assertIsInstance(
                change_collector.differ, TravisPullRequestBuildDiffer)

    def test_collect_file_diffs(self):
        n = 10
        filename = 'file{i}.py'
        with mock_repo() as repo:
            commits = make_mock_commits(repo, n=n, filename=filename)
            contrib_module_path = None
            commit_range = make_commit_range(
                commits[0], commits[-1])

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(self.pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }

            with patch.dict('os.environ', travis_env_vars, clear=True):
                project_path = repo.working_tree_dir
                project = make_mock_project(repo, self.pr_num, project_path,
                                            contrib_module_path)
                change_collector = ChangeCollector(project)
                file_diffs = change_collector._collect_file_diffs()

                # checks on file_diffs
                self.assertEqual(len(file_diffs), n - 1)

                for diff in file_diffs:
                    self.assertEqual(diff.change_type, 'A')
                    self.assertTrue(diff.b_path.startswith('file'))
                    self.assertTrue(diff.b_path.endswith('.py'))

    @unittest.expectedFailure
    def test_categorize_file_diffs(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_collect_features(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_collect_changes(self):
        raise NotImplementedError


class FileChangeValidatorTest(_CommonSetup, unittest.TestCase):

    def test_validation_failure_inadmissible_file_diffs(self):
        path_content = [
            ('readme.txt', None),
            ('src/foo/__init__.py', None),
            ('src/foo/contrib/__init__.py', None),
            ('src/foo/contrib/user_foo/feature_bar.py', None),
            ('src/foo/contrib/user_foo/__init__.py', None),
            ('invalid.py', None),
        ]
        contrib_module_path = 'src/foo/contrib/'
        with mock_file_change_validator(
            path_content, self.pr_num, contrib_module_path
        ) as validator:
            changes = validator.change_collector.collect_changes()
            self.assertEqual(len(changes.file_diffs), 1)
            self.assertEqual(len(changes.candidate_feature_diffs), 0)
            self.assertEqual(len(changes.valid_init_diffs), 0)
            self.assertEqual(len(changes.inadmissible_diffs), 1)
            self.assertEqual(
                changes.inadmissible_diffs[0].b_path, 'invalid.py')

            # TODO
            # self.assertTrue(imported_okay)

            result = validator.validate()
            self.assertFalse(result)

    def test_validation_success(self):
        path_content = [
            ('bob.xml', '<hello>'),
            ('src/foo/__init__.py', None),
            ('src/foo/contrib/__init__.py', None),
            ('src/foo/contrib/user_foo/__init__.py', None),
            ('src/foo/contrib/user_foo/feature_bar.py', None),
        ]
        contrib_module_path = 'src/foo/contrib/'
        with mock_file_change_validator(
            path_content, self.pr_num, contrib_module_path
        ) as validator:
            result = validator.validate()
            self.assertTrue(result)


class FeatureApiValidatorTest(_CommonSetup, unittest.TestCase):

    def test_validation_failure_no_features_found(self):
        path_content = [
            ('readme.txt', None),
            ('src/foo/__init__.py', None),
            ('src/foo/contrib/__init__.py', None),
            ('src/foo/contrib/user_foo/__init__.py', None),
            ('src/foo/contrib/user_foo/feature_bar.py', None),
        ]
        contrib_module_path = 'src/foo/contrib/'
        with mock_feature_api_validator(
            path_content, self.pr_num, contrib_module_path, self.X, self.y
        ) as validator:
            result = validator.validate()
            self.assertFalse(result)

    def test_validation_failure_invalid_feature(self):
        path_content = [
            ('foo.jpg', None),
            ('src/foo/__init__.py', None),
            ('src/foo/contrib/__init__.py', None),
            ('src/foo/contrib/user_foo/__init__.py', None),
            ('src/foo/contrib/user_foo/feature_bar.py',
             self.invalid_feature_str),
        ]
        contrib_module_path = 'src/foo/contrib/'
        with mock_feature_api_validator(
            path_content, self.pr_num, contrib_module_path, self.X, self.y
        ) as validator:
            changes = validator.change_collector.collect_changes()
            self.assertEqual(len(changes.file_diffs), 1)
            self.assertEqual(len(changes.candidate_feature_diffs), 1)
            self.assertEqual(len(changes.valid_init_diffs), 0)
            self.assertEqual(len(changes.inadmissible_diffs), 0)

            # TODO
            # self.assertEqual(len(new_features), 1)
            # self.assertTrue(imported_okay)

            result = validator.validate()
            self.assertFalse(result)

    def test_validation_failure_import_error(self):
        import_error_str = dedent('''
            edf foo():  pass
        ''').strip()
        path_content = [
            ('foo.jpg', None),
            ('src/foo/__init__.py', None),
            ('src/foo/contrib/__init__.py', None),
            ('src/foo/contrib/user_foo/__init__.py', None),
            ('src/foo/contrib/user_foo/feature_baz.py', import_error_str),
        ]
        contrib_module_path = 'src/foo/contrib/'
        with mock_feature_api_validator(
            path_content, self.pr_num, contrib_module_path, self.X, self.y
        ) as validator:
            changes = validator.change_collector.collect_changes()
            self.assertEqual(len(changes.file_diffs), 1)
            self.assertEqual(len(changes.candidate_feature_diffs), 1)
            self.assertEqual(len(changes.valid_init_diffs), 0)
            self.assertEqual(len(changes.inadmissible_diffs), 0)

            # TODO
            # self.assertEqual(len(new_feature_info), 0)
            # self.assertFalse(imported_okay)

            result = validator.validate()
            self.assertFalse(result)



import pkgutil
import types

from funcy import collecting, notnone

from ballet.feature import Feature
from ballet.project import Project
from ballet.util import dfilter
from ballet.util.log import logger

__all__ = (
    'collect_contrib_features',
)


def collect_contrib_features(project_root):
    """Collect contributed features for a project at project_root

    For a project ``foo``, walks modules within the ``foo.features.contrib``
    subpackage. A single object that is an instance of ``ballet.Feature`` is
    imported if present in each module. The resulting ``Feature`` objects are
    collected.

    Args:
        project_root (str, path-like): Path to project root

    Returns:
        List[ballet.Feature]: list of Feature objects
    """
    # TODO Project should require ModuleType
    project = Project(project_root)
    contrib = project._resolve('.features.contrib')
    return _collect_contrib_features(contrib)


@dfilter(notnone)
@collecting
def _collect_contrib_features(module):
    """Collect contributed features from within given module

    Be very careful with untrusted code. The module/package will be
    walked, every submodule will be imported, and all the code therein will be
    executed. But why would you be trying to import from an untrusted package
    anyway?

    Args:
        contrib (module): module (standalone or package) that contains feature
            definitions

    Returns:
        List[Feature]: list of features
    """

    if isinstance(module, types.ModuleType):
        # any module that has a __path__ attribute is also a package
        if hasattr(module, '__path__'):
            yield from _collect_contrib_features_from_package(module)
        else:
            yield _collect_contrib_feature_from_module(module)
    else:
        raise ValueError('Input is not a module')


@collecting
def _collect_contrib_features_from_package(package):
    logger.debug(
        'Walking package path {path} to detect modules...'
        .format(path=package.__path__))

    for importer, modname, _ in pkgutil.walk_packages(
            path=package.__path__,
            prefix=package.__name__ + '.',
            onerror=logger.error):

        try:
            mod = importer.find_module(modname).load_module(modname)
        except ImportError:
            logger.exception(
                'Failed to import module {modname}'
                .format(modname=modname))
            continue

        yield _collect_contrib_feature_from_module(mod)


def _collect_contrib_feature_from_module(mod):
    logger.debug(
        'Trying to import contributed feature from module {modname}...'
        .format(modname=mod.__name__))

    candidates = []
    for attr in dir(mod):
        obj = getattr(mod, attr)
        if isinstance(obj, Feature):
            candidates.append(obj)

    if len(candidates) == 1:
        feature = candidates[0]
        feature.source = mod.__name__
        logger.debug(
            'Imported 1 feature from {modname} from {Feature.__name__} object'
            .format(modname=mod.__name__, Feature=Feature))
        return feature
    elif len(candidates) > 1:
        logger.debug(
            'Found too many {Feature.__name__} objects in module {modname}, '
            'skipping; candidates were {candidates!r}'
            .format(Feature=Feature, modname=mod.__name__,
                    candidates=candidates))
        return None
    else:
        logger.debug(
            'Failed to import anything useful from module {modname}'
            .format(modname=mod.__name__))
        return None



import pathlib
import tempfile
import unittest
from textwrap import dedent

from funcy import contextmanager

from ballet.contrib import _collect_contrib_features
from ballet.util.mod import import_module_at_path


def create_contrib_modules_at_dir(dirname, modcontent, n=1):
    '''Create the a "contrib" directory structure at dirname

    Creates up to mod{n}:

        dirname/
        ├── __init__.py
        ├── mod1/
        |   ├── __init__.py
        |   └── foo1.py
        ├── mod2/
        |   ├── __init__.py
        |   └── foo2.py
        etc.
    '''

    root = pathlib.Path(dirname)
    root.joinpath('__init__.py').touch()
    for i in range(n):
        root.joinpath('mod{i}'.format(i=i)).mkdir()
        root.joinpath('mod{i}'.format(i=i), '__init__.py').touch()

        try:
            modcontent_i = modcontent.format(i=i)
        except KeyError:
            # shouldn't happen
            modcontent_i = modcontent
        modpath_i = str(root.joinpath(
            'mod{i}'.format(i=i), 'foo{i}.py'.format(i=i)))
        with open(modpath_i, 'w') as f:
            f.write(modcontent_i)


class ContribTest(unittest.TestCase):

    def test_collect_contrib_features_stdlib(self):
        # give a nonsense *module*, shouldn't import anything. this is a bad
        # test because it relies on module not defining certain names
        import math
        features = _collect_contrib_features(math)

        # features should be an empty list
        self.assertEqual(len(features), 0)

    def test_collect_contrib_features_thirdparty(self):
        # give a nonsense *package*, shouldn't import anything. this is a bad
        # test because it relies on module not defining certain names
        import funcy
        features = _collect_contrib_features(funcy)
        self.assertEqual(len(features), 0)

    def test_collect_contrib_features_generated(self):
        n = 4
        content = dedent(
            '''
            from ballet import Feature
            from sklearn.preprocessing import StandardScaler
            input = 'col{i}'
            transformer = StandardScaler()
            feature = Feature(input, transformer)
            '''
        ).strip()
        modname = 'contrib_features_generated'
        with self.mock_contrib_module(modname, content, n) as (mod, features):
            self.assertEqual(len(features), n)

    @contextmanager
    def mock_contrib_module(self, modname, content, n):
        with tempfile.TemporaryDirectory() as tmpdir:
            modpath = pathlib.Path(tmpdir).joinpath(modname)
            modpath.mkdir()
            create_contrib_modules_at_dir(modpath, content, n=n)
            mod = import_module_at_path(modname, modpath)
            features = _collect_contrib_features(mod)
            yield mod, features



import pathlib
import warnings
from copy import deepcopy
from enum import Enum
from os import devnull

import numpy as np
import pandas as pd
import sklearn.datasets
from funcy import complement, decorator, lfilter, wraps

from ballet.compat import redirect_stderr, redirect_stdout
from ballet.exc import BalletWarning

RANDOM_STATE = 1754


def asarray2d(a):
    """Cast to 2d array"""
    arr = np.asarray(a)
    if arr.ndim == 1:
        arr = arr.reshape(-1, 1)
    return arr


def get_arr_desc(arr):
    """Get array description, in the form '<array type> <array shape>'"""
    type_ = type(arr).__name__  # see also __qualname__
    shape = getattr(arr, 'shape', None)
    if shape is not None:
        desc = '{type_} {shape}'
    else:
        desc = '{type_} <no shape>'
    return desc.format(type_=type_, shape=shape)


def get_enum_keys(cls):
    return [attr for attr in dir(cls) if not attr.startswith('_')]


def get_enum_values(cls):
    if issubclass(cls, Enum):
        return [getattr(cls, attr).value for attr in get_enum_keys(cls)]
    else:
        return [getattr(cls, attr) for attr in get_enum_keys(cls)]


def indent(text, n=4):
    """Indent each line of text by n spaces"""
    _indent = ' ' * n
    return '\n'.join(_indent + line for line in text.split('\n'))


def make_plural_suffix(obj, suffix='s'):
    if len(obj) != 1:
        return suffix
    else:
        return ''


@decorator
def whether_failures(call):
    """Collects failures and return (success, list_of_failures)"""
    failures = list(call())
    return not failures, failures


def has_nans(obj):
    """Check if obj has any NaNs

    Compatible with different behavior of np.isnan, which sometimes applies
    over all axes (py35, py35) and sometimes does not (py34).
    """
    nans = np.isnan(obj)
    while np.ndim(nans):
        nans = np.any(nans)
    return bool(nans)


@decorator
def dfilter(call, pred):
    """Decorate a callable with a filter that accepts a predicate

    Example::

        >>> @dfilter(lambda x: x >= 0)
        ... def numbers():
        ...     return [-1, 2, 0, -2]
        [2, 0]
    """
    return lfilter(pred, call())


def load_sklearn_df(name):
    method_name = 'load_{name}'.format(name=name)
    method = getattr(sklearn.datasets, method_name)
    data = method()
    X_df = pd.DataFrame(data=data.data, columns=data.feature_names)
    y_df = pd.Series(data.target, name='target')
    return X_df, y_df


@decorator
def quiet(call):
    with open(devnull, 'w') as fnull:
        with redirect_stderr(fnull), redirect_stdout(fnull):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return call()


class DeepcopyMixin:

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result


def one_or_raise(l):
    n = len(l)
    if n == 1:
        return l[0]
    else:
        raise ValueError('Expected exactly 1 element, but got {n}'
                         .format(n=n))


def needs_path(f):
    """Wraps a function that accepts path-like to give it a pathlib.Path"""
    @wraps(f)
    def wrapped(pathlike, *args, **kwargs):
        path = pathlib.Path(pathlike)
        return f(path, *args, **kwargs)
    return wrapped


def warn(msg):
    """Issue a warning message of category BalletWarning"""
    warnings.warn(msg, category=BalletWarning)


@decorator
def raiseifnone(call):
    """Decorate a function to raise a ValueError if result is None"""
    result = call()
    if result is None:
        raise ValueError
    else:
        return result


def falsy(o):
    return isinstance(o, str) and (o.lower() == 'false' or o == '')


truthy = complement(falsy)



from ballet.validation.base import BaseValidator
from ballet.validation.common import ChangeCollector


class ProjectStructureValidator(BaseValidator):

    def __init__(self, project):
        self.change_collector = ChangeCollector(project)

    def validate(self):
        changes = self.change_collector.collect_changes()
        return not changes.inadmissible_diffs



import numbers

import numpy as np
import scipy.stats
from funcy import decorator, suppress
from scipy.special import digamma, gamma
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_consistent_length

from ballet.util import asarray2d
from ballet.util.log import logger

__all__ = (
    'estimate_conditional_information',
    'estimate_entropy',
    'estimate_mutual_information',
)

N_NEIGHBORS = 3   # hyperparameter k from KSG estimator
NEIGHBORS_ALGORITHM = 'auto'
NEIGHBORS_METRIC = 'chebyshev'
DISC_COL_UNIQUE_VAL_THRESH = 0.05


@decorator
def nonnegative(call, name=None):
    """Warn if the function's return value is negative and set it to 0"""
    result = call()
    with suppress(TypeError):
        if result < 0:
            result = 0.0
            # Format a nice log message
            if name is None:
                try:
                    pieces = call._func.__name__.split('_')[1:]
                    name = ''.join(map(str.capitalize, pieces))
                except RuntimeError:
                    name = 'Result'
            logger.warning('%s should be non-negative.', name)
    return result


# Helpers

def _make_neighbors(**kwargs):
    return NearestNeighbors(algorithm=NEIGHBORS_ALGORITHM,
                            metric=NEIGHBORS_METRIC,
                            **kwargs)


def _compute_empirical_probability(x):
    """Compute empirical probability of events in x

    Args:
        x: array-like

    Returns:
        pk: array-like of shape (K,) where where p[k] is the probability of
            event k
        events: array-like of shape (K, m) where each event is a vector of
            length m and there are K unique events
    """
    x = asarray2d(x)
    n, _ = x.shape
    events, counts = np.unique(x, axis=0, return_counts=True)
    pk = counts * 1.0 / n
    return pk, events


def _compute_volume_unit_ball(d, metric=NEIGHBORS_METRIC):
    """Compute volume of a d-dimensional unit ball in R^d with given metric"""
    if metric == 'chebyshev':
        return 1
    elif metric == 'euclidean':
        return np.power(np.pi, d / 2) / gamma(1 + d / 2) / 2**d
    else:
        msg = 'metric {metric} not supported'.format(metric=metric)
        raise ValueError(msg)


def _is_column_disc(col):
    # Heuristics to decide if column is discrete

    # Integer columns are discrete
    if issubclass(col.dtype.type, numbers.Integral):
        return True

    # Real-valued columns that are close to integer values are discrete
    rounding_error = col - col.astype(int)
    if np.allclose(rounding_error, np.zeros(col.size)):
        return True

    # Columns with a small fraction of distinct values are discrete
    uniques = np.unique(col)
    if (uniques.size / col.size) < DISC_COL_UNIQUE_VAL_THRESH:
        return True

    return False


def _is_column_cont(col):
    return not _is_column_disc(col)


def _get_disc_columns(x):
    return np.apply_along_axis(_is_column_disc, 0, x)


# Computing epsilon

def _compute_epsilon(x):
    """Calculate epsilon from KSG Estimator

    Represents twice the distance of each element to its k-th nearest neighbor.

    Args:
        x (array-like): An array with shape (n_samples, n_features)

    Returns:
        array-like: An array with shape (n_samples, 1) representing
            epsilon as described above.

    References:

    .. [1] A. Kraskov, H. Stogbauer and P. Grassberger, "Estimating mutual
           information". Phys. Rev. E 69, 2004.

    """
    k = N_NEIGHBORS
    n = x.shape[0]

    disc_mask = _get_disc_columns(x)
    if np.all(disc_mask):
        # if no continuous columns, there's no point getting epsilon
        return -np.inf
    c = x[:, ~disc_mask]

    nn = _make_neighbors(n_neighbors=k)
    nn.fit(c)
    distances = np.zeros(n)

    # if the kth neighbor is at distance 0, then we are in trouble
    # but we can try the trick of increasing k if we don't use the old
    # value of k sometime later
    while not np.all(distances) and k < n:
        distances, _ = nn.kneighbors(n_neighbors=k)
        distances = distances[:, -1]  # distances to k-nearest neighbor
        k += 1

    return asarray2d(2. * distances)


def _compute_n_points_within_radius_i(nn, x_i, radius_i):
    if x_i.shape[0] != 1 or radius_i.shape[0] != 1:
        raise ValueError

    # adjustment as radius_neighbors would find exact matches
    radius_i = np.nextafter(radius_i, 0)

    ind = nn.radius_neighbors(X=x_i, radius=radius_i, return_distance=False)
    return ind[0].size


def _ithrow(x, i):
    # seem to need to index as x[i:i+1, :] to get a (1,m) row array.
    return x[i:i + 1, :]


def _compute_n_points_within_radius(x, radius):
    """Compute the number of points strictly within some radius

    Note that points lying exactly on the radius are not counted.

    Args:
        x (array-like): data of shape (n_instances, n_features)
        radius (array-like): radius from each point of shape (n_instances, 1)
    """
    check_consistent_length(x, radius)
    n = x.shape[0]

    nn = _make_neighbors()
    nn.fit(x)

    # this will be slow
    nx = np.array([
        _compute_n_points_within_radius_i(nn, _ithrow(x, i), radius[i])
        for i in range(n)
    ])

    return nx


# Entropy estimation

def _estimate_disc_entropy(x):
    r"""Estimate the Shannon entropy of a discrete dataset.

    The Shannon entropy of a discrete random variable :math:`Z` with support
    :math:`\mathbb{Z}` and density :math:`P_Z` is given as

    .. math::
        H(Z) = -\sum_{z \in \mathbb{Z}} P_Z(z) \log(P_Z(z))

    Here, since we do not know :math:`P_Z`, we estimate :math:`\hat{P}_Z`, the
    empirical probability, calculated as the frequency in the dataset x.

    If x's columns logically represent continuous features, it is better to use
    the `_estimate_cont_entropy` function. If you are unsure of which to use,
    `estimate_entropy` can take datasets of mixed discrete and continuous
    functions.

    Args:
        x (array-like): Dataset with shape (n_samples, n_features) or
            (n_samples, )

    Returns:
        float: the dataset entropy.
    """
    x = asarray2d(x)
    pk, _ = _compute_empirical_probability(x)
    return scipy.stats.entropy(pk)


def _estimate_cont_entropy(x, epsilon):
    """Estimate the differential entropy of a continuous dataset.

    Based off the KSG Estimator [1] for a dataset's differential entropy.
    If epsilon is provided, this is a partial estimation of the KSG entropy
    estimator. The bias is cancelled out when computing mutual information.

    The function relies on nonparametric methods based on entropy estimation
    from k-nearest neighbors distances as proposed in [1] and augmented in [2]
    for mutual information estimation.

    If X's columns logically represent discrete features, it is better to use
    the _estimate_disc_entropy function. If you are unsure of which to use,
    _estimate_entropy can take datasets of mixed discrete and continuous
    functions.

    Observe that differential entropy is *not* the "extension" of the
    Shannon entropy and thus it does not exhibit some properties like
    non-negativity (i.e. values below zero are possible).

    Args:
        x (array-like): Dataset with shape (n_samples, n_features) or
            (n_samples, )
        epsilon (array-like): An array with shape (n_samples, 1) that is
            the epsilon used in KSG Estimator. Represents the Chebyshev
            distance from an element to its k-th nearest neighbor in the full
            dataset.

    Returns:
        float: differential entropy of the dataset

    References:

    .. [1] A. Kraskov, H. Stogbauer and P. Grassberger, "Estimating mutual
           information". Phys. Rev. E 69, 2004.

    """
    x = asarray2d(x)
    n, d = x.shape
    nx = _compute_n_points_within_radius(x, epsilon / 2.0)
    c_d = _compute_volume_unit_ball(d)
    return -np.mean(digamma(nx + 1)) + digamma(n) + np.log(c_d) \
        + d * np.mean(np.log(epsilon))


def _estimate_entropy(x, epsilon):
    """Estimate dataset entropy."""
    x = asarray2d(x)
    n, d = x.shape

    # not enough data
    if n <= 1 or d == 0:
        return 0

    disc_mask = _get_disc_columns(x)
    cont_mask = ~disc_mask

    # if all columns are disc, use discrete-specific estimator
    if np.all(disc_mask):
        return _estimate_disc_entropy(x)

    # if all columns are cont, use continuous-specific estimator
    if np.all(cont_mask):
        return _estimate_cont_entropy(x, epsilon)

    # Separate the dataset into discrete and continuous datasets d and c
    d = asarray2d(x[:, disc_mask])
    c = asarray2d(x[:, cont_mask])

    # H(c|d)
    H_c_d = _estimate_conditional_entropy(c, d, epsilon)

    # H(d)
    H_d = _estimate_disc_entropy(d)

    return H_d + H_c_d


def _estimate_conditional_entropy(c, d, epsilon):
    """Estimate H(c|d) where c is continuous and d is discrete"""
    # H(c|d) = \sum_{i} p(d_i) H(c|d=d_i)
    # where we i ranges over unique values of d and c_d_i is the
    # rows of c where d == d_i.
    pk, events = _compute_empirical_probability(d)
    H_c_d = 0.0
    for i, (p_i, d_i) in enumerate(zip(pk, events)):
        mask = np.all(d == d_i, axis=1)
        c_di = c[mask, :]
        # TODO add logging here about small sample size
        epsilon_di = epsilon[mask, :]
        H_c_di = _estimate_cont_entropy(c_di, epsilon_di)
        H_c_d += p_i * H_c_di
    return H_c_d


# Public API

@nonnegative()
def estimate_entropy(x):
    r"""Estimate dataset entropy.

    This function can take datasets of mixed discrete and continuous features,
    and uses a set of heuristics to determine which functions to apply to
    each. Discrete (Shannon) entropy is estimated via the empirical
    probability mass function. Continuous (differential) entropy is
    estimated via the KSG estimator [1].

    Let x be made of continuous features c and discrete features d.
    To deal with both continuous and discrete features, We use the
    following reworking of entropy:

    .. math::
       :nowrap:

       \begin{align}
       H(x) &= H(c,d) \\
            &= H(d) + H(c | d) \\
            &= \sum_{x \in d} p(x) H(c(x)) + H(d),
       \end{align}

    where :math:`c(x)` is a dataset that represents the rows of the continuous
    dataset in the same row as a discrete column with value x in the original
    dataset.

    Args:
        x (array-like): Dataset with shape (n_samples, n_features) or
            (n_samples, )
        epsilon (array-like): An array with shape (n_samples, 1) that is
            the epsilon used in KSG Estimator. Represents the chebyshev
            distance from an element to its k-th nearest neighbor in the full
            dataset.

    Returns:
        float: Dataset entropy of X.

    References:

    .. [1] A. Kraskov, H. Stogbauer and P. Grassberger, "Estimating mutual
           information". Phys. Rev. E 69, 2004.
    """
    x = asarray2d(x)
    epsilon = _compute_epsilon(x)
    return _estimate_entropy(x, epsilon)


@nonnegative()
def estimate_conditional_information(x, y, z):
    r"""Estimate the conditional mutual information of x and y given z

    Conditional mutual information is the mutual information of two datasets,
    given a third:

    .. math::
       I(x;y|z) = H(x,z) + H(y,z) - H(x,y,z) - H(z)

    Where :math:`H(X)` is the Shannon entropy of dataset :math:`X`. For
    continuous datasets, adapts the KSG Estimator [1] for mutual
    information.

    Eq 8 from [1] holds because the epsilon terms cancel out.
    Let :math:`d_x`, represent the dimensionality of the continuous portion of
    x. Then, we see that:

    .. math::
       :nowrap:

       \begin{align}
       d_{xz} + d_{yz} - d_{xyz} - d_z
           &= (d_x + d_z) + (d_y + d_z) - (d_x + d_y + d_z) - d_z \\
           &= 0
       \end{align}

    Args:
        x (array-like): An array with shape (n_samples, n_features_x)
        y (array-like): An array with shape (n_samples, n_features_y)
        z (array-like): An array with shape (n_samples, n_features_z).
            This is the dataset being conditioned on.

    Returns:
        float: conditional mutual information of x and y given z.

    References:

    .. [1] A. Kraskov, H. Stogbauer and P. Grassberger, "Estimating mutual
           information". Phys. Rev. E 69, 2004.
    """
    xz = np.concatenate((x, z), axis=1)
    yz = np.concatenate((y, z), axis=1)
    xyz = np.concatenate((x, y, z), axis=1)

    epsilon = _compute_epsilon(xyz)

    H_xz = _estimate_entropy(xz, epsilon)
    H_yz = _estimate_entropy(yz, epsilon)
    H_xyz = _estimate_entropy(xyz, epsilon)
    H_z = _estimate_entropy(z, epsilon)

    logger.debug('H(X,Z): %s', H_xz)
    logger.debug('H(Y,Z): %s', H_yz)
    logger.debug('H(X,Y,Z): %s', H_xyz)
    logger.debug('H(Z): %s', H_z)

    return H_xz + H_yz - H_xyz - H_z


@nonnegative()
def estimate_mutual_information(x, y):
    r"""Estimate the mutual information of two datasets.

    Mutual information is a measure of dependence between
    two datasets and is calculated as:

    .. math::
       I(x;y) = H(x) + H(y) - H(x,y)

    Where H(x) is the Shannon entropy of x. For continuous datasets,
    adapts the KSG Estimator [1] for mutual information.

    Args:
        x (array-like): An array with shape (n_samples, n_features_x)
        y (array-like): An array with shape (n_samples, n_features_y)

    Returns:
        float: mutual information of x and y

    References:

    .. [1] A. Kraskov, H. Stogbauer and P. Grassberger, "Estimating mutual
      information". Phys. Rev. E 69, 2004.
    """
    xy = np.concatenate((x, y), axis=1)
    epsilon = _compute_epsilon(xy)
    H_x = _estimate_entropy(x, epsilon)
    H_y = _estimate_entropy(y, epsilon)
    H_xy = _estimate_entropy(xy, epsilon)
    return H_x + H_y - H_xy



import logging

import ballet
import ballet.contrib
import ballet.util.mod
import click
from ballet.eng.misc import IdentityTransformer
from ballet.pipeline import BuildResult, FeatureEngineeringPipeline
from ballet.util.io import save_features, save_targets
from stacklog import stacklog

import {{ cookiecutter.package_slug }}
from {{ cookiecutter.package_slug }}.load_data import load_data


logger = logging.getLogger(__name__)


def collect_contrib_features():
    """Get contrib features for this project

    Returns:
        List[ballet.Feature]: list of Feature objects
    """
    return ballet.contrib.collect_contrib_features({{ cookiecutter.package_slug }})


def get_target_encoder():
    """Get encoder for the prediction target

    Returns:
        transformer-like
    """
    return IdentityTransformer()


@stacklog(logger.info, 'Building features and target')
def build(X_df=None, y_df=None):
    """Build features and target

    Args:
        X_df (DataFrame): raw variables
        y_df (DataFrame): raw target

    Returns:
        dict with keys X_df, features, mapper_X, X, y_df, encoder_y, y
    """
    if X_df is None:
        X_df, _ = load_data()
    if y_df is None:
        _, y_df = load_data()

    features = collect_contrib_features()
    mapper_X = FeatureEngineeringPipeline(features)
    X = mapper_X.fit_transform(X_df)

    encoder_y = get_target_encoder()
    y = encoder_y.fit_transform(y_df)

    return BuildResult(X_df=X_df, features=features, mapper_X=mapper_X,
                       y_df=y_df, encoder_y=encoder_y, y=y)


@click.command()
@click.argument('input_dir', type=click.Path(
    exists=True, file_okay=False, readable=True))
@click.argument('output_dir', type=click.Path(file_okay=False))
def main(input_dir, output_dir):
    """Engineer features"""

    import ballet.util.log
    ballet.util.log.enable(logger=logger, level='INFO', echo=False)
    ballet.util.log.enable(logger=ballet.util.log.logger, level='INFO',
                           echo=False)

    X_df, y_df = load_data(input_dir=input_dir)
    out = build()

    mapper_X = out.mapper_X
    encoder_y = out.encoder_y

    X_ft = mapper_X.transform(X_df)
    y_ft = encoder_y.transform(y_df)

    save_features(X_ft, output_dir)
    save_targets(y_ft, output_dir)


if __name__ == '__main__':
    main()



import random

from ballet.util import asarray2d
from ballet.util.log import logger
from ballet.util.testing import seeded
from ballet.validation.base import FeaturePruner, FeaturePruningMixin
from ballet.validation.common import RandomFeaturePerformanceEvaluator
from ballet.validation.entropy import (
    estimate_conditional_information, estimate_entropy)
from ballet.validation.gfssf import (
    LAMBDA_1_ADJUSTMENT, LAMBDA_2_ADJUSTMENT, _compute_lmbdas,
    _compute_threshold, _concat_datasets)


class NoOpPruner(FeaturePruner):
    def prune(self):
        logger.info('Pruning features using {!s}'.format(self))
        return []


class RandomPruner(FeaturePruningMixin, RandomFeaturePerformanceEvaluator):

    def prune(self):
        """With probability p, select a random feature to prune"""
        logger.info('Pruning features using {!s}'.format(self))
        with seeded(self.seed):
            if random.uniform(0, 1) < self.p:
                return [random.choice(self.features)]


CMI_MESSAGE = "Calculating CMI of feature and target cond. on accpt features"


class GFSSFPruner(FeaturePruner):
    """A feature pruning evaluator that uses a modified version of
    GFSSF[1] - specifically, lines 12-13 of agGFSSF.

    Attributes:
        X_df (array-like): The dataset to build features off of.
        y (array-like): A single-column dataset representing the target
            feature.
        features (array-like): an array of ballet Features that have
            already been accepted, to be pruned.
        new_feature (ballet.Feature): the recently accepted Feature.
        lmbda_1 (float): A float used in GFSSF to calculate the information
            threshold. Default is a function of the entropy of y.
        lmbda_2 (float): A float used in GFSSF to calculate the information
            threshold. Default is a function of the entropy of y.

    References:
        [1] H. Li, X. Wu, Z. Li and W. Ding, "Group Feature Selection
            with Streaming Features," 2013 IEEE 13th International
            Conference on Data Mining, Dallas, TX, 2013, pp. 1109-1114.
            doi: 10.1109/ICDM.2013.137

    """

    def __init__(self, *args, lmbda_1=0.0, lmbda_2=0.0):
        super().__init__(*args)
        self.y = asarray2d(self.y)
        if lmbda_1 <= 0:
            lmbda_1 = estimate_entropy(self.y) / LAMBDA_1_ADJUSTMENT
        if lmbda_2 <= 0:
            lmbda_2 = estimate_entropy(self.y) / LAMBDA_2_ADJUSTMENT
        self.lmbda_1 = lmbda_1
        self.lmbda_2 = lmbda_2

    def prune(self):
        feature_dfs_by_src = {}
        for accepted_feature in [self.candidate_feature] + self.features:
            accepted_df = accepted_feature.as_feature_engineering_pipeline(
            ).fit_transform(self.X_df, self.y)
            feature_dfs_by_src[accepted_feature.source] = accepted_df

        lmbda_1, lmbda_2 = _compute_lmbdas(
            self.lmbda_1, self.lmbda_2, feature_dfs_by_src
        )

        logger.info(
            "Pruning features using GFSSF: lambda_1={l1}, lambda_2={l2}"
            .format(l1=lmbda_1, l2=lmbda_2)
        )

        redundant_features = []
        for candidate_feature in self.features:
            candidate_src = candidate_feature.source
            logger.debug("Pruning feature: {}".format(candidate_src))
            candidate_df = feature_dfs_by_src[candidate_src]
            _, n_candidate_cols = candidate_df.shape
            z = _concat_datasets(feature_dfs_by_src, omit=candidate_src)
            logger.debug(CMI_MESSAGE)
            cmi = estimate_conditional_information(candidate_df, self.y, z)

            logger.debug(
                "Conditional Mutual Information Score: {}".format(cmi))
            statistic = cmi
            threshold = _compute_threshold(lmbda_1, lmbda_2, n_candidate_cols)
            logger.debug("Calculated Threshold: {}".format(threshold))
            if statistic >= threshold:
                logger.debug(
                    "Passed, keeping feature: {}".format(candidate_src))
            else:
                logger.debug(
                    "Failed, found redundant feature: {}".format(candidate_src)
                )
                del feature_dfs_by_src[candidate_src]
                redundant_features.append(candidate_feature)
        return redundant_features



import pathlib
import sys
from importlib import import_module

import git
from dynaconf import LazySettings
from funcy import cached_property, fallback, re_find

from ballet.compat import safepath
from ballet.exc import ConfigurationError
from ballet.util import needs_path, raiseifnone
from ballet.util.ci import get_travis_branch, get_travis_pr_num
from ballet.util.git import get_branch, get_pr_num, is_merge_commit
from ballet.util.mod import import_module_at_path

DEFAULT_CONFIG_NAME = 'ballet.yml'
DYNACONF_OPTIONS = {
    'ENVVAR_PREFIX_FOR_DYNACONF': 'BALLET',
    'SETTINGS_FILE_FOR_DYNACONF': DEFAULT_CONFIG_NAME,
    'YAML_LOADER': 'safe_load',
}


config = LazySettings(**DYNACONF_OPTIONS)


@needs_path
def load_config_at_path(path):
    """Load config at exact path

    Args:
        path (path-like): path to config file

    Returns:
        dict: config dict
    """
    if path.exists() and path.is_file():
        options = DYNACONF_OPTIONS.copy()
        options.update({
            'ROOT_PATH_FOR_DYNACONF': safepath(path.parent),
            'SETTINGS_FILE_FOR_DYNACONF': safepath(path.name),
        })
        return LazySettings(**options)
    else:
        raise ConfigurationError(
            'Couldn\'t find ballet.yml config file at {path!s}'
            .format(path=path))


@needs_path
def load_config_in_dir(path):
    """Load config in containing directory

    Args:
        path (path-like): path to containing directory of config file

    Returns:
        dict: config dict
    """
    return load_config_at_path(path.joinpath(DEFAULT_CONFIG_NAME))


def relative_to_contrib(diff, project):
    """Compute relative path of changed file to contrib dir

    Args:
        diff (git.diff.Diff): file diff
        project (Project): project

    Returns:
        Path
    """
    path = pathlib.Path(diff.b_path)
    contrib_path = project.config.get('contrib.module_path')
    return path.relative_to(contrib_path)


@needs_path
def make_feature_path(contrib_dir, username, featurename):
    return contrib_dir.joinpath(
        'user_{}'.format(username), 'feature_{}.py'.format(featurename))


def detect_github_username(project):
    """Detect github username

    Looks in the following order:
    1. github.user git config variable
    2. git remote origin
    3. $USER
    4. 'username'
    """
    @raiseifnone
    def get_config_variable():
        return project.repo.config_reader().get_value(
            'github', 'user', default=None)

    @raiseifnone
    def get_remote():
        url = list(project.repo.remote('origin').urls)[0]
        # protocol:user/repo, i.e. 'git@github.com:HDI-Project/ballet'
        return re_find(r'.+:(.+)/.+', url)

    @raiseifnone
    def get_user_env():
        return sys.environ.get('USER')

    def get_default():
        return 'username'

    return fallback(
        get_config_variable, get_remote, get_user_env, (get_default, ())
    )


class Project:
    """Encapsulate information on a ballet project

    This is a utility class mostly useful for easy access to the project's
    information from within the ballet.validation package.

    In addition to the defined methods and properties, the following functions
    of the project can be accessed as attributes of a class instance, where
    ``prj`` refers to the python module of the underlying ballet project:
    - ``load_data`` (``prj.load_data.load_data``)
    - ``build`` (``prj.features.build``)
    - ``collect_contrib_features`` (``prj.features.collect_contrib_features``)

    Args:
        package (ModuleType): python package representing imported ballet
            project
    """

    attr_map = {
        'load_data': ('.load_data', 'load_data'),
        'build': ('.features', 'build'),
        'collect_contrib_features': ('.features', 'collect_contrib_features')
    }

    def __init__(self, package):
        self.package = package

    @cached_property
    def config(self):
        return load_config_in_dir(self.path)

    @classmethod
    def from_path(cls, path):
        """Create a Project instance from an fs path to the containing dir

        Args:
            path (PathLike): path to directory that contains the
                project
        """
        path = pathlib.Path(path)
        config = load_config_in_dir(path)
        package_slug = config.get('project.package_slug')
        package = import_module_at_path(package_slug,
                                        path.joinpath('src', package_slug))
        return cls(package)

    @classmethod
    def from_cwd(cls):
        """Create a Project instance by searching up from cwd

        Recursively searches for the ballet configuration file at the
        current working directory and parent directories, stopping when it
        reaches a file system boundary.

        Raises:
            ConfigurationError: couldn't find the configuration file
        """
        path = pathlib.Path.cwd()
        while path.exists() and not path.is_mount():
            try:
                return Project.from_path(path)
            except ConfigurationError:
                path = path.parent

        raise ConfigurationError('Couldn\'t create Project instance')

    def _resolve(self, modname, attr=None):
        module = import_module(modname, package=self.package.__name__)
        if attr is not None:
            return getattr(module, attr)
        else:
            return module

    @property
    def pr_num(self):
        """Return the PR number or None if not on a PR"""
        result = get_pr_num(repo=self.repo)
        if result is None:
            result = get_travis_pr_num()
        return result

    @property
    def on_pr(self):
        """Return whether the project has a source tree on a PR"""
        return self.pr_num is not None

    @property
    def branch(self):
        """Return whether the project is on master branch"""
        result = get_branch(repo=self.repo)
        if result is None:
            result = get_travis_branch()
        return result

    @property
    def on_master(self):
        return self.branch == 'master'

    @property
    def on_master_after_merge(self):
        """Check the repo HEAD is on master after a merge commit

        Checks for two qualities of the current project:
        1. The project repo's head is the master branch
        2. The project repo's head commit is a merge commit.

        Note that fast-forward style merges will not cause the second condition
        to evaluate to true.
        """

        return self.on_master and is_merge_commit(self.repo.head.commit)

    @property
    def path(self):
        """Return the project path (aka project root)

        If ``package.__file__`` is ``/foo/src/foo/__init__.py``,
        then project.path
        should be ``/foo``.
        """
        return pathlib.Path(self.package.__file__).resolve().parents[2]

    @property
    def repo(self):
        """Return a git.Repo object corresponding to this project"""
        return git.Repo(self.path, search_parent_directories=True)

    def __getattr__(self, attr):
        if attr in Project.attr_map:
            return self._resolve(*Project.attr_map[attr])
        else:
            return object.__getattribute__(self, attr)



import funcy
import numpy as np
import pandas as pd
from scipy.special import boxcox1p
from scipy.stats import skew
from sklearn.preprocessing import FunctionTransformer
from sklearn.utils.validation import check_is_fitted

from ballet.eng.base import BaseTransformer, ConditionalTransformer
from ballet.util import get_arr_desc

__all__ = [
    'IdentityTransformer',
    'BoxCoxTransformer',
    'ValueReplacer',
    'NamedFramer',
]


class IdentityTransformer(FunctionTransformer):
    """Simple transformer that passes through its input"""

    def __init__(self):
        super().__init__(func=funcy.identity, inverse_func=funcy.identity,
                         validate=False, check_inverse=False)


class BoxCoxTransformer(ConditionalTransformer):
    """Conditionally apply the Box-Cox transformation

    In the fit stage, determines which variables (columns) have absolute skew
    above ``threshold``. In the transform stage, applies the Box-Cox
    transformation of 1+x to each variable selected previously.

    Args:
        threshold: skew threshold.
        lmbda (float, default=0.0): Power parameter of the Box-Cox transform.

    See also:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.boxcox1p.html
    """

    def __init__(self, threshold, lmbda=0):
        def condition(X):
            return abs(skew(X)) > threshold

        def transform(X):
            return boxcox1p(X, lmbda)

        super().__init__(condition, transform)


class ValueReplacer(BaseTransformer):
    """Replace instances of some value with some replacement

    Args:
        value: value to replace (checked by equality testing)
        replacement: replacement
    """

    def __init__(self, value, replacement):
        super().__init__()
        self.value = value
        self.replacement = replacement

    def transform(self, X, **transform_kwargs):
        X = X.copy()
        mask = X == self.value
        X[mask] = self.replacement
        return X


class NamedFramer(BaseTransformer):
    """Convert object to named 1d DataFrame

    If transformation is successful, the resulting object is a DataFrame with a
    ``name`` attribute as given.

    Args:
        name: name for resulting DataFrame
    """

    def __init__(self, name):
        super().__init__()
        self.name = name

    def transform(self, X, **transform_kwargs):
        msg = "Couldn't convert object {} to named 1d DataFrame."
        if isinstance(X, pd.Index):
            return X.to_series().to_frame(name=self.name)
        elif isinstance(X, pd.Series):
            return X.to_frame(name=self.name)
        elif isinstance(X, pd.DataFrame):
            if X.shape[1] == 1:
                X = X.copy()
                X.columns = [self.name]
                return X
            else:
                raise ValueError(msg.format(get_arr_desc(X)))
        elif isinstance(X, np.ndarray):
            if X.ndim == 1:
                return pd.DataFrame(data=X.reshape(-1, 1), columns=[self.name])
            elif X.ndim == 2 and X.shape[1] == 1:
                return pd.DataFrame(data=X, columns=[self.name])
            else:
                raise ValueError(msg.format(get_arr_desc(X)))

        raise TypeError(msg.format(get_arr_desc(X)))


class NullTransformer(BaseTransformer):

    def transform(self, X, **transform_kwargs):
        n = np.size(X, 0)
        return np.empty((n, 0))


class ComputedValueTransformer(BaseTransformer):

    def __init__(self, func, pass_y=False):
        self.func = func
        self.pass_y = pass_y

    def fit(self, X, y=None, **fit_kwargs):
        if self.pass_y:
            self.value_ = self.func(X, y=y)
        else:
            self.value_ = self.func(X)
        self.dtype_ = np.dtype(type(self.value_))

    def transform(self, X, **transform_kwargs):
        check_is_fitted(self, ['value_', 'dtype_'])
        return np.full_like(X, self.value_, dtype=self.dtype)



import io
from copy import deepcopy

import dill as pickle
import numpy as np
from funcy import all, all_fn, isa, iterable

from ballet.feature import Feature
from ballet.validation.base import BaseCheck


def _get_one_row(*args):
    return tuple(
        obj.iloc[0:1]
        for obj in args
    )


class FeatureApiCheck(BaseCheck):
    """Base class for implementing new Feature API checks

    Args:
        X_df (array-like): X dataframe
        y_df (array-like): y dataframe
    """

    def __init__(self, X_df, y_df):
        self.X = X_df
        self.y = y_df


class IsFeatureCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the object is an instance of ballet.Feature"""
        assert isinstance(feature, Feature)


class HasCorrectInputTypeCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature's `input` is a str or Iterable[str]"""
        input = feature.input
        is_str = isa(str)
        is_nested_str = all_fn(
            iterable, lambda x: all(is_str, x))
        assert is_str(input) or is_nested_str(input)


class HasTransformerInterfaceCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature has a fit/transform/fit_tranform interface"""
        assert hasattr(feature.transformer, 'fit')
        assert hasattr(feature.transformer, 'transform')
        assert hasattr(feature.transformer, 'fit_transform')


class CanMakeMapperCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature can be converted to a FEP"""
        feature.as_feature_engineering_pipeline()


class CanFitCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that fit can be called on reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit(self.X, y=self.y)


class CanFitOneRowCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that fit can be called on one row of reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        x, y = _get_one_row(self.X, self.y)
        mapper.fit(x, y=y)


class CanTransformCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that transform can be called on reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit(self.X, y=self.y)
        mapper.transform(self.X)


class CanTransformOneRowCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that transform can be called on one row of reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit(self.X, y=self.y)
        x, = _get_one_row(self.X)
        mapper.transform(x)


class CanFitTransformCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that fit_transform can be called on reference data"""
        mapper = feature.as_feature_engineering_pipeline()
        mapper.fit_transform(self.X, y=self.y)


class HasCorrectOutputDimensionsCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the dimensions of the transformed data are correct

        For input X, an n x p array, a n x q array should be produced,
        where q is the number of features produced by the logical feature.
        """
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        assert self.X.shape[0] == X.shape[0]


class CanDeepcopyCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature can be deepcopied

        This is needed for execution of the overall transformation pipeline
        """
        deepcopy(feature)


class CanPickleCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the feature can be pickled

        This is needed for saving the pipeline to disk
        """
        try:
            buf = io.BytesIO()
            pickle.dump(feature, buf, protocol=pickle.HIGHEST_PROTOCOL)
            buf.seek(0)
            new_feature = pickle.load(buf)
            assert new_feature is not None
            assert isinstance(new_feature, Feature)
        finally:
            buf.close()


class NoMissingValuesCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the output of the transformer has no missing values"""
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        assert not np.any(np.isnan(X))


class NoInfiniteValuesCheck(FeatureApiCheck):

    def check(self, feature):
        """Check that the output of the transformer has no non-finite values"""
        mapper = feature.as_feature_engineering_pipeline()
        X = mapper.fit_transform(self.X, y=self.y)
        assert not np.any(np.isinf(X))



import os
from subprocess import CalledProcessError, check_call
from textwrap import dedent
from types import ModuleType
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from ballet.compat import safepath
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.pipeline import FeatureEngineeringPipeline
from ballet.project import make_feature_path
from ballet.templating import start_new_feature
from ballet.util.code import get_source
from ballet.util.git import make_commit_range, switch_to_new_branch
from ballet.util.log import logger
from ballet.util.mod import import_module_at_path, modname_to_relpath
from tests.util import load_regression_data


def submit_feature(repo, contrib_dir, username, featurename, new_feature_str):
    feature_path = make_feature_path(contrib_dir, username, featurename)
    cc_kwargs = {
        'extra_context': {'username': username, 'featurename': featurename},
        'no_input': True,
    }
    result = start_new_feature(contrib_dir=contrib_dir, **cc_kwargs)
    with feature_path.open('w') as f:
        f.write(new_feature_str)

    added_files = [str(fn) for (fn, kind) in result if kind == 'file']
    repo.index.add(added_files)
    repo.index.commit('Add {} feature'.format(feature_path))


def make_feature_str(input):
    return dedent("""
        from ballet import Feature
        from ballet.eng.misc import IdentityTransformer
        input = {input!r}
        transformer = IdentityTransformer()
        feature = Feature(input, transformer)
    """.format(input=input)).strip()


@pytest.mark.skip
@pytest.mark.slow
def test_validation_end_to_end(quickstart):
    project = quickstart.project
    modname = 'foo'
    base = project.path
    repo = project.repo

    def _import(modname):
        relpath = modname_to_relpath(modname,
                                     project_root=base,
                                     add_init=False)
        abspath = base.joinpath(relpath)
        return import_module_at_path(modname, abspath)

    foo = _import('foo')
    assert isinstance(foo, ModuleType)

    foo_features = _import('foo.features')
    assert isinstance(foo_features, ModuleType)

    collect_contrib_features = foo_features.collect_contrib_features
    features = collect_contrib_features()
    assert len(features) == 0

    # first providing a mock feature, call build
    with patch.object(
        foo_features, 'collect_contrib_features',
        return_value=[Feature(input='A_1', transformer=IdentityTransformer())]
    ):
        X_df = pd.util.testing.makeCustomDataframe(5, 2)
        X_df.columns = ['A_0', 'A_1']
        out = foo_features.build(X_df=X_df, y_df=[])
        assert np.shape(out.X) == (5, 1)
        assert isinstance(out.mapper_X, FeatureEngineeringPipeline)

    # write a new version of foo.load_data.load_data
    new_load_data_str = get_source(load_regression_data)

    p = base.joinpath(modname, 'load_data.py')
    with p.open('w') as f:
        f.write(new_load_data_str)

    # commit changes
    repo.index.add([str(p)])
    repo.index.commit('Load mock regression dataset')

    # call different validation routines
    def call_validate_all(pr=None):
        envvars = {
            'TRAVIS_BUILD_DIR': repo.working_tree_dir,
        }
        if pr is None:
            envvars['TRAVIS_PULL_REQUEST'] = 'false'
            envvars['TRAVIS_COMMIT_RANGE'] = make_commit_range(
                repo.commit('HEAD@{-1}').hexsha, repo.commit('HEAD').hexsha)
            envvars['TRAVIS_PULL_REQUEST_BRANCH'] = ''
            envvars['TRAVIS_BRANCH'] = repo.heads.master.name
        else:
            envvars['TRAVIS_PULL_REQUEST'] = str(pr)
            envvars['TRAVIS_COMMIT_RANGE'] = make_commit_range(
                repo.heads.master.name,
                repo.commit('pull/{pr}'.format(pr=pr)).hexsha)

        with patch.dict(os.environ, envvars):
            cmd = 'ballet validate -A'
            check_call(cmd, cwd=safepath(base), env=os.environ)

    call_validate_all()

    # branch to a fake PR and write a new feature
    contrib_dir = base.joinpath(modname, 'features', 'contrib')
    logger.info('Switching to pull request 1, User Bob, Feature A')
    switch_to_new_branch(repo, 'pull/1')
    new_feature_str = make_feature_str('A_0')
    username = 'bob'
    featurename = 'A_0'
    submit_feature(repo, contrib_dir, username, featurename, new_feature_str)

    # call different validation routines
    logger.info('Validating pull request 1, User Bob, Feature A')
    call_validate_all(pr=1)

    # merge PR with master
    logger.info('Merging into master')
    repo.git.checkout('master')
    repo.git.merge('pull/1', no_ff=True)

    # call different validation routines
    logger.info('Validating after merge')
    call_validate_all()

    # write another new feature
    logger.info('Switching to pull request 2, User Charlie, Feature Z_1')
    switch_to_new_branch(repo, 'pull/2')
    new_feature_str = make_feature_str('Z_1')
    username = 'charlie'
    featurename = 'Z_1'
    submit_feature(repo, contrib_dir, username, featurename, new_feature_str)

    # if we expect this feature to fail
    with pytest.raises(CalledProcessError):
        logger.info('Validating pull request 2, User Charlie, Feature Z_1')
        call_validate_all(pr=2)

    # write another new feature - redudancy
    repo.git.checkout('master')
    switch_to_new_branch(repo, 'pull/3')
    new_feature_str = make_feature_str('A_0')
    username = 'charlie'
    featurename = 'A_0'
    submit_feature(repo, contrib_dir, username, featurename, new_feature_str)

    with pytest.raises(CalledProcessError):
        call_validate_all(pr=3)



from ballet import Feature


input = None  # TODO - str or list of str
transformer = None  # TODO - function, transformer-like, or list thereof
name = None  # TODO - str
feature = Feature(input, transformer, name)



import sys


# sklearn compatibility
try:
    from sklearn.impute import SimpleImputer
except ImportError:
    import sklearn.preprocessing
    SimpleImputer = sklearn.preprocessing.Imputer


if sys.version_info < (3, 6):
    safepath = str
else:
    from funcy import identity as _identity
    safepath = _identity


try:
    from os import PathLike
except ImportError:
    from pathlib import Path
    PathLike = (Path, )


# redirect_stderr new in 3.5
from contextlib import redirect_stdout  # noqa F401
try:
    from contextlib import redirect_stderr
except ImportError:
    from contextlib import contextmanager
    @contextmanager
    def redirect_stderr(f):
        oldstream = sys.stderr
        try:
            sys.stderr = f
            yield f
        finally:
            sys.stderr = oldstream


# black compatibility - only installable for python 3.6+
try:
    import black
except ImportError:
    black = None




class BalletWarning(UserWarning):
    """Base warning for ballet"""
    pass


class BalletError(Exception):
    """Base error for ballet"""
    pass


class UnexpectedTravisEnvironmentError(BalletError):
    """The environment within Travis CI testing was unexpected"""
    pass


class UnsuccessfulInputConversionError(BalletError):
    """Input-type conversion for execution within pipeline was unsuccessful"""
    pass


class ConfigurationError(BalletError):
    """BalletError in configuration of the ballet project"""
    pass


class FeatureValidationError(BalletError):
    """Base error for feature validation routine"""
    pass


class FeatureRejected(FeatureValidationError):
    """The candidate feature was rejected"""
    pass


class InvalidProjectStructure(FeatureValidationError):
    """An invalid change to the ballet project structure was introduced"""
    pass


class InvalidFeatureApi(FeatureValidationError):
    """The candidate feature had an invalid API"""
    pass


class FeatureCollectionError(BalletError):
    """Error in collecting Feature instances from some source"""
    pass


class NoFeaturesCollectedError(FeatureCollectionError):
    """Expected to collect some features but did not find any"""
    pass


class SkippedValidationTest(BalletError):
    """The validation test was skipped"""
    pass



import pathlib
import tempfile

import funcy as fy
from cookiecutter.main import cookiecutter as _cookiecutter

from ballet.compat import PathLike
from ballet.project import Project, detect_github_username
from ballet.util.fs import pwalk, synctree
from ballet.util.log import logger
from ballet.validation.project_structure.checks import (
    FEATURE_MODULE_NAME_REGEX, SUBPACKAGE_NAME_REGEX)

TEMPLATES_PATH = pathlib.Path(__file__).resolve().parent.joinpath('templates')
FEATURE_TEMPLATE_PATH = TEMPLATES_PATH.joinpath('feature_template')
PROJECT_TEMPLATE_PATH = TEMPLATES_PATH.joinpath('project_template')


def _stringify_path(obj):
    return str(obj) if isinstance(obj, PathLike) else obj


@fy.wraps(_cookiecutter)
def cookiecutter(*args, **kwargs):
    """Call cookiecutter.main.cookiecutter after stringifying paths

    Return:
        str: project directory path
    """
    args = fy.walk(_stringify_path, args)
    kwargs = fy.walk_values(_stringify_path, kwargs)
    return _cookiecutter(*args, **kwargs)


def render_project_template(project_template_path=None, **cc_kwargs):
    """Generate a ballet project according to the project template

    Args:
        project_template_path (str): path to specific project template
        **cc_kwargs: options for the cookiecutter template
    """
    if project_template_path is None:
        project_template_path = PROJECT_TEMPLATE_PATH
    return cookiecutter(project_template_path, **cc_kwargs)


def render_feature_template(**cc_kwargs):
    """Create a stub for a new feature

    Args:
        **cc_kwargs: options for the cookiecutter template
    """
    feature_template_path = FEATURE_TEMPLATE_PATH
    return cookiecutter(feature_template_path, **cc_kwargs)


def _fail_if_feature_exists(dst):
    subpackage_name, feature_name = str(dst.parent), str(dst.name)
    if (
        dst.is_file()
        and fy.re_test(SUBPACKAGE_NAME_REGEX, subpackage_name)
        and fy.re_test(FEATURE_MODULE_NAME_REGEX, feature_name)
    ):
        raise FileExistsError(
            'The feature already exists here: {dst}'
            .format(dst=dst))


def start_new_feature(contrib_dir=None, **cc_kwargs):
    """Start a new feature within a ballet project

    By default, will prompt the user for input using cookiecutter's input
    interface.

    Renders the feature template into a temporary directory, then copies the
    feature files into the proper path within the contrib directory.

    Args:
        contrib_dir: directory under which to place contributed features
        **cc_kwargs: options for the cookiecutter template

    Raises:
        ballet.exc.BalletError: the new feature has the same name as an
            existing one
    """
    if contrib_dir is None:
        project = Project.from_path(pathlib.Path.cwd().resolve())
        contrib_dir = project.config.get('contrib.module_path')

    # inject default username into context
    default_username = detect_github_username(project)
    cc_kwargs.setdefault('extra_context', {})
    cc_kwargs['extra_context'].update({'_default_username': default_username})

    with tempfile.TemporaryDirectory() as tempdir:
        # render feature template
        output_dir = tempdir
        cc_kwargs['output_dir'] = output_dir
        rendered_dir = render_feature_template(**cc_kwargs)

        # clean pyc files from rendered dir
        for path in pwalk(rendered_dir, topdown=False):
            if path.suffix == '.pyc':
                path.unlink()
            if path.name == '__pycache__':
                with fy.suppress(OSError):
                    path.rmdir()

        # copy into contrib dir
        src = rendered_dir
        dst = contrib_dir
        result = synctree(src, dst, onexist=_fail_if_feature_exists)

    _log_start_new_feature_success(result)

    return result


def _log_start_new_feature_success(result):
    logger.info('Start new feature successful.')
    for (name, kind) in result:
        if kind == 'file' and '__init__' not in str(name):
            relname = pathlib.Path(name).relative_to(pathlib.Path.cwd())
            logger.info('Created {}'.format(relname))



import random

from ballet.util import asarray2d
from ballet.util.log import logger
from ballet.util.testing import seeded
from ballet.validation.base import FeatureAcceptanceMixin, FeatureAccepter
from ballet.validation.common import RandomFeaturePerformanceEvaluator
from ballet.validation.entropy import (
    estimate_conditional_information, estimate_entropy)
from ballet.validation.gfssf import (
    LAMBDA_1_ADJUSTMENT, LAMBDA_2_ADJUSTMENT, _compute_lmbdas,
    _compute_threshold, _concat_datasets)


class NeverAccepter(FeatureAccepter):

    def judge(self):
        logger.info('Judging feature using {}'.format(self))
        return False


class RandomAccepter(FeatureAcceptanceMixin,
                     RandomFeaturePerformanceEvaluator):

    def judge(self):
        """Accept feature with probability p"""
        logger.info('Judging feature using {}'.format(self))
        with seeded(self.seed):
            return random.uniform(0, 1) < self.p


class AlwaysAccepter(FeatureAccepter):
    def judge(self):
        logger.info('Judging feature using {}'.format(self))
        return True


class GFSSFAccepter(FeatureAccepter):
    """A feature acceptance evaluator that uses a modified version of
    GFSSF[1] - specifically, lines 1 - 8 of agGFSSF where we do not
    remove accepted but redundant features on line 8.

    Attributes:
        X_df (array-like): The dataset to build features off of.
        y (array-like): A single-column dataset representing the target
            feature.
        features (array-like): an array of ballet Features that have
            already been accepted.
        lmbda_1 (float): A float used in GFSSF to calculate the information
            threshold. Default is a function of the entropy of y.
        lmbda_2 (float): A float used in GFSSF to calculate the information
            threshold. Default is a function of the entropy of y.

    References:
        [1] H. Li, X. Wu, Z. Li and W. Ding, "Group Feature Selection
            with Streaming Features," 2013 IEEE 13th International
            Conference on Data Mining, Dallas, TX, 2013, pp. 1109-1114.
            doi: 10.1109/ICDM.2013.137

    """

    def __init__(self, *args, lmbda_1=0., lmbda_2=0.):
        super().__init__(*args)
        self.y = asarray2d(self.y)
        if lmbda_1 <= 0:
            lmbda_1 = estimate_entropy(self.y) / LAMBDA_1_ADJUSTMENT
        if lmbda_2 <= 0:
            lmbda_2 = estimate_entropy(self.y) / LAMBDA_2_ADJUSTMENT
        self.lmbda_1 = lmbda_1
        self.lmbda_2 = lmbda_2

    def __str__(self):
        return '{str}: lmbda_1={lmbda_1}, lmbda_2={lmbda_2}'.format(
            str=super().__str__(),
            lmbda_1=self.lmbda_1,
            lmbda_2=self.lmbda_2)

    def judge(self):
        logger.info('Judging Feature using {}'.format(self))
        feature_dfs_by_src = {}
        for feature in [self.candidate_feature] + self.features:
            feature_df = (feature
                          .as_feature_engineering_pipeline()
                          .fit_transform(self.X_df, self.y))
            feature_dfs_by_src[feature.source] = feature_df

        candidate_source = self.candidate_feature.source
        candidate_df = feature_dfs_by_src[candidate_source]
        n_samples, n_candidate_cols = feature_df.shape

        lmbda_1, lmbda_2 = _compute_lmbdas(
            self.lmbda_1, self.lmbda_2, feature_dfs_by_src)

        logger.info(
            'Candidate Feature Shape: {}'.format(candidate_df.shape))
        omit_in_test = [''] + [f.source for f in self.features]
        for omit in omit_in_test:
            logger.debug(
                'Testing with omitted feature: {}'.format(
                    omit or 'None'))
            z = _concat_datasets(
                feature_dfs_by_src, n_samples, [
                    candidate_source, omit])
            logger.debug('Calculating CMI of candidate feature:')
            cmi = estimate_conditional_information(candidate_df, self.y, z)
            logger.debug(
                'Conditional Mutual Information Score: {}'.format(cmi))
            cmi_omit = 0
            n_omit_cols = 0
            if omit:
                omit_df = feature_dfs_by_src[omit]
                _, n_omit_cols = omit_df.shape
                logger.debug(
                    'Calculating CMI of ommitted feature:'
                )
                cmi_omit = estimate_conditional_information(
                    omit_df, self.y, z)
                logger.debug('Omitted CMI Score: {}'.format(cmi_omit))
                logger.debug('Omitted Feature Shape: {}'.format(omit_df.shape))
            statistic = cmi - cmi_omit
            threshold = _compute_threshold(
                lmbda_1, lmbda_2, n_candidate_cols, n_omit_cols)
            logger.debug('Calculated Threshold: {}'.format(threshold))
            if statistic >= threshold:
                logger.debug(
                    'Succeeded while omitting feature: {}'.format(
                        omit or 'None'))
                return True
        return False



import pytest
from cookiecutter.utils import work_in

from ballet.compat import safepath


@pytest.mark.slow
def test_quickstart_install(quickstart, virtualenv):
    # This is an annoying test because the project template should specify
    # as installation dependency the most recent tagged version of ballet,
    # but this will not necessarily have been released on PyPI
    # TODO: figure out the right way to mange this
    d = quickstart.tempdir.joinpath(quickstart.project_slug).absolute()
    with work_in(safepath(d)):
        # cmd = 'cd "{d!s}" && make install'.format(d=d)
        cmd = 'echo okay'
        virtualenv.run(cmd, capture=True)



import unittest

import numpy as np
import pandas as pd
import sklearn.base
from sklearn.impute import SimpleImputer

import ballet.eng.base
import ballet.exc
from ballet.util.testing import ArrayLikeEqualityTestingMixin


class TestBase(ArrayLikeEqualityTestingMixin, unittest.TestCase):

    def setUp(self):
        pass

    def test_no_fit_mixin(self):
        class _Foo(ballet.eng.base.NoFitMixin):
            pass
        a = _Foo()

        self.assertTrue(hasattr(a, 'fit'))

        # method should exist
        a.fit('X')
        a.fit('X', y=None)

    def test_base_transformer(self):
        a = ballet.eng.base.BaseTransformer()

        self.assertIsInstance(a, sklearn.base.BaseEstimator)
        self.assertTrue(hasattr(a, 'fit'))

    def test_simple_function_transformer(self):
        def func(x): return x + 5
        data = np.arange(30)

        trans = ballet.eng.base.SimpleFunctionTransformer(func)
        trans.fit(data)
        data_trans = trans.transform(data)
        data_func = func(data)

        self.assertArrayEqual(data_trans, data_func)

    def test_grouped_function_transformer(self):
        df = pd.DataFrame(
            data={
                'country': ['USA', 'USA', 'USA', 'Canada', 'Fiji'],
                'year': [2001, 2002, 2003, 2001, 2001],
                'length': [1, 2, 3, 4, 5],
                'width': [1.0, 1.0, 7.5, 9.0, 11.0],
            }
        ).set_index(['country', 'year']).sort_index()

        # with groupby kwargs, produces a df
        func = np.sum
        trans = ballet.eng.base.GroupedFunctionTransformer(
            func, groupby_kwargs={'level': 'country'})
        trans.fit(df)
        result = trans.transform(df)
        expected_result = df.groupby(level='country').apply(func)
        self.assertFrameEqual(result, expected_result)

        # without groupby kwargs, produces a series
        func = np.min
        trans = ballet.eng.base.GroupedFunctionTransformer(func)
        trans.fit(df)
        result = trans.transform(df)
        expected_result = df.pipe(func)
        self.assertSeriesEqual(result, expected_result)


class GroupwiseTransformerTest(
    ArrayLikeEqualityTestingMixin, unittest.TestCase
):

    def setUp(self):
        self.X_tr = pd.DataFrame(
            data={
                'name': ['A', 'A', 'A', 'B', 'B', 'C', 'C'],
                'year': [2001, 2002, 2003, 2001, 2002, 2001, 2003],
                'value': [1, 2, np.nan, 4, 4, 5, np.nan],
                'size': [3, 5, 5, 5, 5, np.nan, 4],
            }
        ).set_index(['name', 'year']).sort_index()

        self.X_te = pd.DataFrame(
            data={
                'name': ['A', 'B', 'C'],
                'year': [2004, 2004, 2004],
                'value': [np.nan, np.nan, np.nan],
                'size': [4, 1, np.nan],
            }
        ).set_index(['name', 'year']).sort_index()

        self.groupby_kwargs = {'level': 'name'}

        # mean-impute within groups
        self.individual_transformer = SimpleImputer()
        self.trans = ballet.eng.base.GroupwiseTransformer(
            self.individual_transformer,
            groupby_kwargs=self.groupby_kwargs,
            column_selection=['value'],
        )

    def test_can_fit(self):
        self.trans.fit(self.X_tr)

    def test_can_transform(self):
        self.trans.fit(self.X_tr)

        result_tr = self.trans.transform(self.X_tr)
        expected_tr = self.X_tr.copy()
        expected_tr['value'] = np.array([1, 2, 1.5, 4, 4, 5, 5])
        expected_tr = expected_tr.drop('size', axis=1)
        self.assertFrameEqual(result_tr, expected_tr)

        result_te = self.trans.transform(self.X_te)
        expected_te = self.X_te.copy()
        expected_te['value'] = np.array([1.5, 4, 5])
        expected_te = expected_te.drop('size', axis=1)
        self.assertFrameEqual(result_te, expected_te)

    def test_raise_on_new_group(self):
        trans = ballet.eng.base.GroupwiseTransformer(
            self.individual_transformer,
            groupby_kwargs=self.groupby_kwargs,
            handle_unknown='error'
        )

        trans.fit(self.X_tr)

        X_te = self.X_te.copy().reset_index()
        X_te.loc[0, 'name'] = 'Z'  # new group
        X_te = X_te.set_index(['name', 'year'])

        with self.assertRaises(ballet.exc.BalletError):
            trans.transform(X_te)

    def test_ignore_on_new_group(self):
        trans = ballet.eng.base.GroupwiseTransformer(
            self.individual_transformer,
            groupby_kwargs=self.groupby_kwargs,
            handle_unknown='ignore'
        )

        trans.fit(self.X_tr)

        X_te = self.X_te.copy().reset_index()
        X_te.loc[0, 'name'] = 'Z'  # new group
        X_te = X_te.set_index(['name', 'year'])

        result = trans.transform(X_te)

        # the first group, Z, is new, and values are passed through, so such
        # be nan
        expected = X_te.copy()
        expected['value'] = np.array([np.nan, 4.0, 5.0])
        expected['size'] = np.array([4.0, 1.0, 4.0])

        self.assertFrameEqual(result, expected)

    def test_raise_on_transform_error(self):
        exc = Exception

        class TransformErrorTransformer(ballet.eng.base.BaseTransformer):
            def transform(self, X, **transform_kwargs):
                raise exc

        individual_transformer = TransformErrorTransformer()
        trans = ballet.eng.base.GroupwiseTransformer(
            individual_transformer,
            groupby_kwargs=self.groupby_kwargs,
            handle_error='error',
        )

        trans.fit(self.X_tr)

        with self.assertRaises(exc):
            trans.transform(self.X_tr)

    def test_ignore_on_transform_error(self):
        exc = Exception

        class TransformErrorTransformer(ballet.eng.base.BaseTransformer):
            def transform(self, X, **transform_kwargs):
                raise exc

        individual_transformer = TransformErrorTransformer()
        trans = ballet.eng.base.GroupwiseTransformer(
            individual_transformer,
            groupby_kwargs=self.groupby_kwargs,
            handle_error='ignore',
        )

        trans.fit(self.X_tr)

        result_tr = trans.transform(self.X_tr)
        expected_tr = self.X_tr

        self.assertFrameEqual(result_tr, expected_tr)

        result_te = trans.transform(self.X_te)
        expected_te = self.X_te
        self.assertFrameEqual(result_te, expected_te)



from inspect import signature

from funcy import cached_property

import ballet.pipeline
from ballet.transformer import make_robust_transformer

__all__ = ('Feature', )


class Feature:
    """A logical feature

    Conceptually, a logical feature is a learned function that maps raw
    variables in one data instance to a vector of feature values. A logical
    feature can produce either a scalar feature value for each instance or a
    vector of feature values, as in the case of an embedding technique like PCA
    or the one-hot encoding of a categorical variable.

    Args:
        input (str, Collection[str]): required columns from the input
            dataframe needed for the transformation
        transformer (transformer-like, List[transformer-like]): transformer, or
            sequence of transformers. A "transformer" is an instance of a class
            that provides a fit/transform-style learned transformation.
            Alternately, a callable can be provided, either by itself or in a
            list, in which case it will be converted into a
            ``SimpleFunctionTransformer`` for convenience.
        name (str, optional): name of the feature
        description (str, optional): description of the feature
        output (str, list[str]): ordered sequence of names of features
            produced by this transformer
        source (path-like): the source file in which this feature was defined
        options (dict): options
    """

    def __init__(self, input, transformer, name=None, description=None,
                 output=None, source=None, options=None):
        self.input = input
        self.transformer = make_robust_transformer(transformer)
        self.name = name
        self.description = description
        self.output = output  # unused
        self.source = source
        self.options = options if options is not None else {}

    @cached_property
    def _init_attr_list(self):
        return list(signature(self.__init__).parameters)

    def __repr__(self):
        indent = 4
        attr_list = self._init_attr_list
        attrs_str = ',\n'.join(
            '{indent}{attr_name}={attr_val!s}'.format(
                indent=' ' * indent,
                attr_name=attr,
                attr_val=getattr(self, attr)
            ) for attr in attr_list
        )
        return '{clsname}(\n{attrs_str}\n)'.format(
            clsname=type(self).__name__, attrs_str=attrs_str)

    def as_input_transformer_tuple(self):
        """Return an tuple for passing to DataFrameMapper constructor"""
        return self.input, self.transformer, {'alias': self.output}

    def as_feature_engineering_pipeline(self):
        """Return standalone FeatureEngineeringPipeline with this feature"""
        return ballet.pipeline.FeatureEngineeringPipeline(self)
